# -*- coding: utf-8 -*-
# ===========================================================================
# Parallel features processing using multi-core CPU and multiprocessing
# Copyright 2016-2017 TrungNT
# ===========================================================================
from __future__ import print_function, division, absolute_import

import re
import os
import sys
import wave
import time
import random
import shutil
import warnings
from numbers import Number
from multiprocessing import Pool, cpu_count, Process, Queue
from six import add_metaclass, string_types
from six.moves import zip, zip_longest, range, cPickle
from abc import ABCMeta, abstractmethod, abstractproperty

from collections import defaultdict, Mapping

import numpy as np

from sklearn.pipeline import Pipeline

from odin.utils.mpi import MPI
from odin.utils import (Progbar, as_tuple, get_all_files, ctext,
                        get_tempdir, is_string, batching,
                        add_notification, defaultdictkey,
                        stdio, get_stdio_path, wprint,
                        add_notification, flatten_list,
                        get_formatted_datetime)
from odin.fuel import Dataset, MmapDict, MmapData
from odin.preprocessing.base import Extractor, ExtractorSignal

_default_module = re.compile(r"__.*__")

# ===========================================================================
# PCA calculation
# ===========================================================================
def calculate_pca(dataset, feat_name='auto', batch_size=1234, override=False):
  """ Using parallel MiniBatchPCA to do PCA for multiple features
  at once.

  """
  # TODO: add different pca prefix (e.g. pca_full_mspec, pca_sami_mspec)
  # add reading data from indices also
  # ====== check input dataset ====== #
  own_dataset = True
  if is_string(dataset) and os.path.isdir(dataset):
    dataset = Dataset(dataset, read_only=True)
  elif isinstance(dataset, Dataset):
    own_dataset = False
  elif isinstance(dataset, FeatureProcessor):
    dataset = Dataset(dataset.path, read_only=True)
  else:
    raise ValueError("Cannot acquire Dataset from input: %s" %
                     str(dataset))
  # ====== extract all feat_name ====== #
  if is_string(feat_name) and feat_name == 'auto':
    feat_name = []
    for k in dataset.keys():
      X = dataset[k]
      if hasattr(X, 'ndim') and X.ndim == 2 and X.shape[-1] > 1:
        feat_name.append(k)
  else:
    feat_name = [name
                 for name in as_tuple(feat_name, t=str)
                 if name in dataset]
  # ====== load PCA ====== #
  from odin.ml import MiniBatchPCA
  # init PCA
  nb_samples = 0
  for feat in feat_name:
    nb_samples += dataset[feat].shape[0]
  # ====== prepare MPI PCA ====== #
  add_notification("Selected features for PCA: " +
      ctext(', '.join(feat_name), 'yellow'))

  def map_pca(name):
    X = dataset[name]
    # found exist pca model
    if 'pca_' + feat in dataset and not override:
      pca = dataset['pca_' + feat]
    # create new PCA
    else:
      pca = MiniBatchPCA(n_components=None, whiten=False,
                         copy=True, batch_size=None)
    # No shuffling make iter much faster
    for x in X.set_batch(batch_size=batch_size, seed=None, shuffle_level=0):
      pca.partial_fit(x)
      yield x.shape[0]
    # save PCA model
    with open(os.path.join(dataset.path, 'pca_' + name), 'wb') as f:
      cPickle.dump(pca, f, protocol=cPickle.HIGHEST_PROTOCOL)
    # finish return feature name
    yield name
  mpi = MPI(jobs=feat_name, func=map_pca,
            ncpu=None, batch=1, hwm=1234,
            backend='python')
  # ====== running the MPI ====== #
  remain_features = list(feat_name)
  finished_features = []
  prog = Progbar(target=nb_samples, print_summary=True, print_report=True,
                 name='PCA')
  for n in mpi:
    if is_string(n):
      remain_features.remove(n)
      finished_features.append(n)
    else:
      prog['Remain'] = ', '.join(remain_features)
      prog['Finished'] = ', '.join(finished_features)
      prog.add(n)
  # ====== return ====== #
  if own_dataset:
    dataset.close()


# ===========================================================================
# Helper
# ===========================================================================
def _escape_file_name(file_name):
  file_name = file_name.replace('/', '_')
  file_name = file_name.replace(':', '_')
  return file_name


def _special_cases(X, feat_name, file_name, ds, path):
  """ Same special for checking the integrity of the features """
  if feat_name == 'raw':
    from odin.preprocessing.speech import save
    sr = ds['sr'][file_name]
    if '.wav' not in file_name:
      file_name += '.wav'
    save(os.path.join(path, _escape_file_name(file_name)),
         X.astype('float32'), sr=sr)
  elif feat_name == 'spec':
    from odin.preprocessing.speech import (SpectraExtractor, STFTExtractor, Power2Db,
                                          _extract_frame_step_length, save)
    from odin.preprocessing.signal import ispec
    sr = ds['sr'][file_name]
    extractor = [i for _, i in ds['pipeline'].steps
                 if isinstance(i, SpectraExtractor) or isinstance(i, STFTExtractor)][0]
    frame_length, step_length = _extract_frame_step_length(sr,
        extractor.frame_length, extractor.step_length)
    raw = ispec(X,
                frame_length=frame_length, step_length=step_length,
                window=extractor.window, padding=extractor.padding,
                db=extractor.log if hasattr(extractor, 'log') else
                any(isinstance(i, Power2Db) for i in ds['pipeline'].steps))
    file_name += '-ispec.wav'
    save(os.path.join(path, _escape_file_name(file_name)),
         raw.astype('float32'), sr=sr)


def validate_features(ds_or_processor, path, nb_samples=25,
                      override=False, seed=1234, fig_width=4):
  # TODO: add PCA visualization
  # TODO: update to match new indices style
  def logger(title, tag, check):
    check = bool(check)
    text_color = 'yellow' if check else 'red'
    print(ctext('   *', 'cyan'),
          ctext(str(title), text_color),
          ctext(str(tag), 'magenta'),
          ctext("✓", text_color) if check else ctext("✗", text_color))
  import matplotlib
  matplotlib.use('Agg')
  from odin.visual import plot_save, plot_multiple_features
  # ====== check path to dataset ====== #
  should_close_ds = True
  if isinstance(ds_or_processor, FeatureProcessor):
    ds = Dataset(ds_or_processor.path, read_only=True)
  elif is_string(ds_or_processor):
    ds = Dataset(ds_or_processor, read_only=True)
  elif isinstance(ds_or_processor, Dataset):
    ds = ds_or_processor
    should_close_ds = False
  else:
    raise ValueError("`ds` can be None, string, or Dataset. No "
                     "support for given input type: %s" % str(type(ds)))
  print(ctext('Validating dataset:', 'yellow'), '"%s"' % ds.path)
  # ====== extract the config of the dataset ====== #
  if 'config' not in ds:
    raise RuntimeError("The `Dataset` must be generated by `FeatureProcessor` "
                       "which must contain `config` MmapDict of extracted "
                       "features configuration.")
  # config = ds['config']
  # pipeline = ds['pipeline']
  # ====== output path ====== #
  path = str(path)
  if not os.path.exists(path):
    os.mkdir(path)
  elif override:
    if os.path.isfile(path):
      os.remove(path)
    else:
      shutil.rmtree(path)
    os.mkdir(path)
  else:
    raise ValueError("`path`=%s exists, cannot override." % path)
  prev_stdio = get_stdio_path()
  stdio(path=os.path.join(path, 'log.txt'))
  nb_samples = int(nb_samples)
  # ====== get all features ====== #
  # [(name, dtype, statistic-able), ...]
  all_keys = [k for k in ds.keys() if k not in ('config', 'pipeline')]
  # store all features (included the features in external_indices
  all_features = []
  # the external indices can be: indices_mfcc_bnf
  external_indices = flatten_list([k.split('_')[1:] for k in all_keys
                                   if 'indices' in k and k != 'indices'])
  # ====== checking indices ====== #
  main_indices = {name: (start, end)
                  for name, (start, end) in ds['indices'].items()}
  for ids_name in (k for k in all_keys if 'indices' in k):
    ids = sorted([(name, start, end)
                  for name, (start, end) in ds[ids_name].items()],
                 key=lambda x: x[1])
    for prev, now in zip(ids, ids[1:]):
      assert prev[2] == now[1], "Zero length in indices"
      assert prev[2] - prev[1] > 0, "Zero length in indices"
      assert now[2] - now[1] > 0, "Zero length in indices"
    # final length match length of Data
    if ids_name != 'indices':
      for feat_name in ids_name.split('_')[1:]:
        assert now[-1] == len(ds[feat_name]), \
            "Indices and data length mismatch, indices:'%s' feat:'%s'" % \
            (ids_name, feat_name)
        all_features.append(feat_name)
    else:
      for feat_name in all_keys:
        if feat_name not in external_indices and \
        'sum1' != feat_name[-4:] and 'sum2' != feat_name[-4:] and \
        'mean' != feat_name[-4:] and 'std' != feat_name[-3:] and \
        isinstance(ds[feat_name], MmapData):
          assert now[-1] == len(ds[feat_name]), \
          "Length of indices and actual data mismatch, " + ids_name + ':' + feat_name
          all_features.append(feat_name)
    # logging
    logger("Checked all:", ids_name, True)
  # ====== check all dictionary types ====== #
  for name in all_keys:
    if isinstance(ds[name], MmapDict) and 'indices' not in name:
      data = ds[name]
      # special cases
      if name == 'sr':
        checking_func = lambda x: x > 0 # for sr
      else:
        checking_func = lambda x: True
      # check
      for key, val in data.items():
        assert key in main_indices, \
        "Dictionary with name:'%s' has key not found in indices." % name
        assert checking_func(val)
      logger("Checked dictionary: ", name, True)
  # ====== checking each type of data ====== #
  # get all stats name
  all_stats = defaultdict(list)
  for k in all_keys:
    if 'sum1' == k[-4:] or 'sum2' == k[-4:] or \
    'mean' == k[-4:] or 'std' == k[-3:]:
      all_stats[k[:-4].split('_')[0]].append(k)
  # get all pca name
  all_pca = {i: i + '_pca' for i in all_features
             if i + '_pca' in ds}
  # checking one-by-one numpy.ndarray features array
  for feat_name in all_features:
    dtype = str(ds[feat_name].dtype)
    # checking all data
    indices = ds.find_prefix(feat_name, 'indices')
    prog = Progbar(target=len(indices), interval=0.1,
                   print_report=True,
                   name='Checking: %s(%s)' % (feat_name, dtype))
    # start iterating over all data file
    fail_test = False
    for file_name, (start, end) in indices:
      dat = ds[feat_name][start:end]
      # No NaN value
      if np.any(np.isnan(dat)):
        logger("NaN values", file_name + ':' + feat_name, False)
        fail_test = True
      # not all value closed to zeros
      if np.all(np.isclose(dat, 0.)):
        logger("All-closed-zeros values", file_name + ':' + feat_name,
               False)
        fail_test = True
      prog['Name'] = file_name
      prog.add(1)
    if not fail_test:
      logger("Check data incredibility for: ", feat_name, True)
    # checking statistics
    if feat_name in all_stats:
      fail_test = False
      for stat_name in all_stats[feat_name]:
        X = ds[stat_name]
        if X.ndim >= 1:
          X = X[:]
        if np.any(np.isnan(X)):
          logger("NaN values", feat_name + ':' + stat_name, False)
          fail_test = True
        if np.all(np.isclose(X, 0.)):
          logger("All-closed-zeros values", feat_name + ':' + stat_name,
                 False)
          fail_test = True
      if not fail_test:
        logger("Check statistics for: ", feat_name, True)
    # check PCA
    if feat_name in all_pca:
      pca = ds[all_pca[feat_name]]
      n = ds[feat_name].shape[0]
      nb_feats = ds[feat_name].shape[-1]
      fail_test = False
      # performing PCA on random samples
      for i in range(nb_samples):
        start = np.random.randint(0, n - nb_samples - 1)
        X = pca.transform(
            ds[feat_name][start:(start + nb_samples)],
            n_components=max(nb_feats // 2, 1))
        if np.any(np.isnan(X)):
          logger("NaN values in PCA", feat_name, False)
          fail_test = True
          break
        if np.all(np.isclose(X, 0.)):
          logger("All-closed-zeros values in PCA", feat_name, False)
          fail_test = True
          break
      if not fail_test:
        logger("Check PCA for: ", feat_name, True)
  # ====== Do sampling ====== #
  np.random.seed(seed) # seed for reproceducible
  all_samples = np.random.choice(list(ds['indices'].keys()),
                                 size=nb_samples,
                                 replace=False)
  # plotting all samples
  for sample_id, file_name in enumerate(all_samples):
    X = {}
    for feat_name in all_features:
      start, end = ds.find_prefix(feat_name, 'indices')[file_name]
      feat = ds[feat_name][start:end]
      X[feat_name] = feat
      # some special handling
      try:
        _special_cases(X=feat, feat_name=feat_name, file_name=file_name,
                       ds=ds, path=path)
      except Exception as e:
        logger("Special case error: %s" % str(e),
               file_name + ':' + feat_name, False)
    plot_multiple_features(X, title=file_name, fig_width=fig_width)
    figure_path = os.path.join(path, '%s.pdf' % _escape_file_name(file_name))
    plot_save(figure_path, log=False, clear_all=True)
    logger("Sample figure saved at: ", figure_path, True)
  # plotting the statistic
  figure_path = os.path.join(path, 'stats.pdf')
  for feat_name, stat_name in all_stats.items():
    X = {name: ds[name][:]
         for name in stat_name
         if ds[name].ndim >= 1}
    if len(X) > 0:
      plot_multiple_features(X, title=feat_name, fig_width=fig_width)
  plot_save(figure_path, log=False, clear_all=True)
  logger("Stats figure save at: ", figure_path, True)
  logger("All reports at folder: ", os.path.abspath(path), True)
  # ====== cleaning ====== #
  stdio(path=prev_stdio)
  if should_close_ds:
    ds.close()

# ===========================================================================
# Features Processor
# ===========================================================================
def _check_logpath(log_path):
  main_path, ext = os.path.splitext(log_path)
  main_path = main_path.split('.')
  try:
    current_log_index = int(main_path[-1])
    main_path = main_path[:-1]
  except ValueError:
    current_log_index = 0
  main_path = '.'.join(main_path)
  # ====== increase log index until found a new file ====== #
  while True:
    path = main_path + '.' + str(current_log_index) + ext
    if not os.path.exists(path):
      break
    current_log_index += 1
  return main_path + '.' + str(current_log_index) + ext

class FeatureProcessor(object):

  """ FeatureProcessor

  Parameters
  ----------
  jobs: list, or tuple
      list of jobs for processing, keep the jobs specification simple
      additional information can be introduced via custom Extractor.

  extractor: sklearn.Pipeline, odin.preprocessing.Extractor
      list of extactor for creating a Pipeline

  path: str
      path to a folder for saving output Dataset

  n_cache: float or int (> 0)
      pass

  ncpu: int (>0)
      number of Processes will be used to parallel the processor

  override : bool (default: False)
      if output folder already exist, deleted old features if
      `override=True`

  identifier : string (default: 'name')
      extractor will return a dictionary, `identifier` is the key
      of the entry that will be used to identify (or distinguish)
      between files (or images, or samples).
      Note this key must exist in the returned dictionary,
      otherwise, RuntimeError will be raised

  log_path : string
      path to log file

  stop_on_failure : bool (default: False)
      There are two types or error during running a FeatureProcessor.

      The first type is handled by `odin.preprocessing.base.Extractor`
      and return `odin.preprocessing.base.ExtractorSignal` which
      contained instruction for handling the Exception.

      The second type is non-handled by the `Extractor`, the
      `FeatureProcessor` will automatically handle this Exception
      getting the traceback and write everything to the log file
      before decided to continue or terminating the process.

      if True, terminate the processor if non-handled Exception
      appeared.
  """

  def __init__(self, jobs, path, extractor,
               n_cache=0.12, ncpu=1, override=False,
               identifier='name',
               log_path=None,
               stop_on_failure=False):
    super(FeatureProcessor, self).__init__()
    # ====== check outpath ====== #
    path = os.path.abspath(str(path))
    if os.path.isfile(path):
      raise ValueError("`path` must be path to a directory, but found a "
                       "path to file.")
    # check override
    if os.path.exists(path) and override:
      wprint("Remove existed Dataset at path: %s" % path)
      for i in os.listdir(path):
        i = os.path.join(path, i)
        if os.path.isdir(i): # remove folder
          shutil.rmtree(i)
        else: # remove file
          os.remove(i)
    # set path and name
    self.path = path
    # ====== check jobs ====== #
    if not isinstance(jobs, (tuple, list, np.ndarray)):
      raise ValueError("Provided `jobs` must be instance of tuple, list or ndarray.")
    if isinstance(jobs, np.ndarray):
      jobs = jobs.tolist()
    self.jobs = tuple(jobs)
    # ====== check multiprocessing ====== #
    if ncpu is None: # auto select number of CPU
      ncpu = min(len(jobs), cpu_count() - 1)
    ncpu = int(ncpu)
    if ncpu <= 0 or n_cache <= 0:
      raise ValueError('`ncpu` and `n_cache` must be greater than 0, but '
                       'given values ncpu=%d n_cache=%f' % (ncpu, n_cache))
    self.n_cpu = ncpu
    self.n_cache = n_cache
    # ====== internal control for feature processor ====== #
    if isinstance(extractor, Pipeline):
      pass
    elif isinstance(extractor, (tuple, list)):
      steps = [('%s_%d' % (e.__class__.__name__, i), e)
               for i, e in enumerate(extractor)]
      extractor = Pipeline(steps=steps)
    elif isinstance(extractor, Mapping):
      steps = [(str(n), e) for n, e in extractor.items()]
      extractor = Pipeline(steps=steps)
    elif isinstance(extractor, Extractor):
      extractor = Pipeline(
          steps=[(extractor.__class__.__name__, extractor)])
    self.extractor = extractor
    # ====== check identifier and log path ====== #
    self._identifier = str(identifier)
    if log_path is None:
      log_path = os.path.join(self.path, 'log.txt')
    else:
      log_path = str(log_path)
    self._log_path = _check_logpath(log_path)
    # ====== others ====== #
    self.config = {}
    self._error_log = []
    self.stop_on_failure = bool(stop_on_failure)

  @property
  def identifier(self):
    return self._identifier

  @property
  def error_log(self):
    return list(self._error_log)

  # ==================== debugging ==================== #
  def __str__(self):
    s = ctext('============= FeatureProcessor: %s =============' % self.path, 'yellow') + '\n'
    padding = '  '
    # ====== basic info ====== #
    s += '- Jobs: ' + ctext(len(self.jobs), 'cyan') + '\n'
    s += '- #CPU: ' + ctext(self.n_cpu, 'cyan') + '\n'
    s += '- #Cache: ' + ctext(self.n_cache, 'cyan') + '\n'
    # ====== print pipeline ====== #
    s += ctext("* Pipeline:", 'yellow') + '\n'
    for _, extractor in self.extractor.steps:
      for line in str(extractor).split('\n'):
        s += padding + ' ' + line + '\n'
    # ====== print config ====== #
    s += ctext("* Configurations:", 'yellow') + '\n'
    for i, j in self.config.items():
      s += padding + str(i) + ' : ' + str(j) + '\n'
    return s

  def __repr__(self):
    return self.__str__()

  # ==================== Abstract properties ==================== #
  def run(self):
    njobs = len(self.jobs)
    dataset = Dataset(self.path)
    if self.n_cache <= 1:
      cache_limit = max(2, int(0.12 * njobs))
    else:
      cache_limit = int(self.n_cache)
    # ====== indices ====== #
    databases = defaultdictkey(lambda key:
        MmapDict(path=os.path.join(dataset.path, key), cache_size=10000,
                 read_only=False))
    last_start = defaultdict(int)
    # ====== statistic ====== #
    # load old statistics
    stats = defaultdict(lambda: [0, 0]) # name -> (sum1, sum2)
    for key in dataset.keys():
      if 'sum1' == key[-4]:
        stats[key[:-4]][0] = dataset[key][:]
      elif 'sum2' == key[-4:]:
        stats[key[:-4]][1] = dataset[key][:]
    # all data are cached for periodically flushed
    cache = defaultdict(list)
    n_processed = [0] # store the value as reference

    # ====== helper ====== #
    def flush_feature(feat_name, X_cached):
      if len(X_cached) > 0:
        X_cached = np.concatenate(X_cached, 0)
        # flush data
        if feat_name in dataset:
          dataset[feat_name].append(X_cached)
        else:
          dataset[(feat_name, 'memmap')] = X_cached

    # ====== repeated for each result returned ====== #
    def post_processing(result):
      # search for file name
      if self.identifier not in result:
        raise RuntimeError(
            "Cannot find identifier '%s' in returned dictionary" % self.identifier)
      file_name = result[self.identifier]
      # invalid file_name
      if not is_string(file_name):
        raise RuntimeError("Cannot find file name in returned features "
            "list, the file name can be specified in key: 'name', 'path' "
            "and the type of the value must be string. All available "
            "keys are: %s" % str(result.keys()))
      # store all new indices
      # mapping [X.shape[0]] -> [feat_name, feat_name, ...]
      all_indices = {}
      # processing
      for feat_name, X in result.items():
        # some invalid feat_name
        if feat_name in ('config', 'pipeline', 'sum1', 'sum2'):
          raise RuntimeError("Returned features' name cannot be one "
                             "of the following: 'config', 'pipeline', 'sum1', 'sum2'.")
        # ignore some feat_name
        if feat_name in ('name'):
          continue
        # if numpy ndarray, save to MmapData
        if isinstance(X, np.ndarray) or \
        'sum1' == feat_name[-4:] or \
        'sum2' == feat_name[-4:]:
          # save statistics instead
          if 'sum1' == feat_name[-4:]:
            stats[feat_name[:-4]][0] += X
          elif 'sum2' == feat_name[-4:]:
            stats[feat_name[:-4]][1] += X
          # save features array
          else:
            all_indices[feat_name] = X.shape[0]
            # cache data, only if we have more than 0 sample
            if X.shape[0] > 0:
              cache[feat_name].append(X)
        # else all other kind of data save to MmapDict
        else:
          databases[feat_name][file_name] = X
        # remove data
        del X
      # ====== update indices ====== #
      if len(all_indices) > 0:
        for feat_name, n in all_indices.items():
          ids_name = 'indices_%s' % feat_name
          databases[ids_name][file_name] = (last_start[ids_name],
                                            last_start[ids_name] + n)
          last_start[ids_name] += n
      # ====== flush cache ====== #
      n_processed[0] += 1
      if n_processed[0] % cache_limit == 0: # 12 + 8
        for feat_name, X_cached in cache.items():
          flush_feature(feat_name, X_cached)
        cache.clear()
      # ====== update progress ====== #
      return file_name

    # ====== mapping function ====== #
    def _map_func(dat):
      try:
        ret = self.extractor.transform(dat)
      except Exception as e: # Non-handled exception
        ret = '\n========\n'
        ret += 'Time  : `%s`\n' % str(get_formatted_datetime(only_number=False))
        ret += 'Error : `%s`\n' % str(e)
        ret += 'Input : `%s`\n' % str(dat)
        import traceback
        etype, value, tb = sys.exc_info()
        for line in traceback.TracebackException(
                type(value), value, tb, limit=None).format(chain=True):
          ret += line
      return ret
    # ====== processing ====== #
    mpi = MPI(jobs=self.jobs,
              func=_map_func,
              ncpu=self.n_cpu,
              batch=1,
              hwm=self.n_cpu * 3,
              backend='python')
    # initialize
    prog = Progbar(target=njobs, name=self.path,
                   interval=0.12, print_report=True, print_summary=True)
    start_time = time.time()
    last_time = time.time()
    last_count = 0
    with open(self._log_path, 'w') as flog:
      # writing the log head
      flog.write('============================\n')
      flog.write('Start Time : %s\n' % get_formatted_datetime(only_number=False))
      flog.write('Outpath    : %s\n' % self.path)
      flog.write('Extractor  : %s\n' % '->'.join([s[-1].__class__.__name__
                                                  for s in self.extractor.steps]))
      flog.write('#Jobs      : %d\n' % njobs)
      flog.write('#CPU       : %d\n' % self.n_cpu)
      flog.write('#Cache     : %d\n' % cache_limit)
      flog.write('============================\n')
      flog.flush()
      # start processing the file list
      for count, result in enumerate(mpi):
        # Non-handled exception
        if isinstance(result, string_types):
          flog.write(result)
          flog.flush()
          self._error_log.append(result)
          if self.stop_on_failure:
            raise RuntimeError(result)
        # some error might happened
        elif isinstance(result, ExtractorSignal):
          flog.write(str(result)); flog.flush()
          if result.action == 'error':
            prog.add_notification(str(result))
            raise RuntimeError("ExtractorSignal requests terminating processor!")
          elif result.action == 'warn':
            prog.add_notification(str(result))
          elif result.action == 'ignore':
            self._error_log.append(result)
          else:
            raise RuntimeError("Unknown action from ExtractorSignal: %s" % result.action)
          prog['File'] = '%-48s' % result.message[:48]
        # otherwise, no error happened, do post-processing
        else:
          name = post_processing(result)
          prog['File'] = '%-48s' % str(name)[:48]
        # update progress
        prog.add(1)
        # manually write to external log file
        if (count + 1) % max(1, int(0.01 * njobs)) == 0:
          curr_time = time.time()
          elap = curr_time - start_time
          avg_speed = (count + 1) / elap
          cur_speed = (count + 1 - last_count) / (curr_time - last_time)
          avg_est = (njobs - count - 1) / avg_speed
          cur_est = (njobs - count - 1) / cur_speed
          flog.write('[%s] Processed: %d(files)   Remain: %d(files)   Elap.: %.2f(secs)\n'
                     '   Avg.Spd: %.2f(obj/sec)  Avg.Est.: %.2f(secs)\n'
                     '   Cur.Spd: %.2f(obj/sec)  Cur.Est.: %.2f(secs)\n' %
                     (get_formatted_datetime(only_number=False),
                      count + 1, njobs - count - 1, elap,
                      avg_speed, avg_est,
                      cur_speed, cur_est))
          flog.flush()
          last_time = curr_time
          last_count = count + 1
    # ====== end, flush the last time ====== #
    for feat_name, X_cached in cache.items():
      flush_feature(feat_name, X_cached)
    cache.clear()
    cache = None
    dataset.flush()
    prog.add_notification("Flushed all data to disk")
    # ====== saving indices ====== #
    for name, db in databases.items():
      db.flush(save_all=True)
      db_size = len(db)
      db.close()
      prog.add_notification('Flush MmapDict "%s" to disk, size: %s' %
                            (ctext(name, 'yellow'),
                             ctext(str(db_size), 'yellow')))

    # ====== save mean and std ====== #
    def save_mean_std(sum1, sum2, name):
      N = dataset[name.split('_')[0]].shape[0]
      mean = sum1 / N
      std = np.sqrt(sum2 / N - np.power(mean, 2))
      if np.any(np.isnan(mean)):
        wprint('Mean contains NaN, name: %s' % name)
      if np.any(np.isnan(std)):
        wprint('Std contains NaN, name: %s' % name)
      dataset[name + 'sum1'] = sum1
      dataset[name + 'sum2'] = sum2
      dataset[name + 'mean'] = mean
      dataset[name + 'std'] = std
    # save all stats
    if len(stats) > 0:
      for feat_name, (sum1, sum2) in stats.items():
        save_mean_std(sum1, sum2, feat_name)
        prog.add_notification('Saved statistics of: %s, shape: %s' %
                              (ctext(feat_name.split('_')[0], 'yellow'),
                               ctext(str(sum1.shape), 'yellow')))
    # ====== dataset flush() ====== #
    dataset.flush()
    dataset.close()
    # ====== saving the extractor ====== #
    # not good idea to save the extractor all the time
    # pipeline_path = os.path.join(dataset.path, 'pipeline')
    # with open(pipeline_path, 'wb') as f:
    #   cPickle.dump(self.extractor, f, protocol=2)
    # prog.add_notification("Saved Extractor pipeline at: %s" %
    #                       ctext(pipeline_path, 'yellow'))
    # ====== saving the configuration ====== #
    config_path = os.path.join(dataset.path, 'config')
    config = MmapDict(config_path)
    config['__configuration_time__'] = time.time()
    config['__processor__'] = self.path
    for i in dir(self):
      if _default_module.match(i) is not None:
        continue
      j = getattr(self, i)
      if isinstance(j, (Number, string_types, bool)):
        config[i] = j
    config.flush(save_all=True)
    self.config = {i: j
                   for i, j in config}
    config.close()
    prog.add_notification("Saved configuration at: %s" %
                          ctext(config_path, 'yellow'))
    # ====== final notification ====== #
    prog.add_notification("Closed all dataset.")
    prog.add_notification("Dataset at path: %s" % ctext(dataset.path, 'yellow'))



from __future__ import absolute_import, division, print_function

from contextlib import contextmanager

import tensorflow as tf
from tensorflow.python.platform import tf_logging as logging


@contextmanager
def suppress_logging(level=logging.ERROR):
  curr_log = logging.get_verbosity()
  logging.set_verbosity(level)
  yield logging
  logging.set_verbosity(curr_log)



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import math
import random

from numbers import Number
from itertools import chain
from collections import defaultdict, Iterator, OrderedDict, Mapping

import numpy as np

from odin.maths import interp
from odin.utils import as_tuple, flatten_list, ctext, batching

def prior2weights(prior, exponential=False,
                  min_value=0.1, max_value=None,
                  norm=False):
  """
  Parameters
  ----------
  prior: numpy.ndarray [nb_classes,]
      probabilty values of each classes prior,
      sum of all prior must be equal to 1.
  exponential: bool
  min_value: bool
      minimum value for the class with highest prior
  max_value: bool
      maximum value for the class with smalles prior
  norm: bool
      if True, normalize output weights to sum up to 1.
  """
  # idea is the one with highest prior equal to 1.
  # and all other classes is the ratio to this prior
  prior = np.array(prior).ravel()
  # make sure everything sum to 1 (probability values)
  prior = prior / np.sum(prior)
  zero_ids = [i for i, j in enumerate(prior) if j == 0]
  nonzero_prior = np.array([j for i, j in enumerate(prior) if j != 0])
  prior = 1. / nonzero_prior * np.max(nonzero_prior)
  if exponential:
    prior = sorted([(i, p) for i, p in enumerate(prior)],
                   key=lambda x: x[-1], reverse=False)
    alpha = interp.expIn(n=len(prior), power=10)
    prior = {i: a * p for a, (i, p) in zip(alpha, prior)}
    prior = np.array([prior[i] for i in range(len(prior))]) + 1
  # ====== rescale everything within max_value ====== #
  if min_value is not None and max_value is not None:
    min_value = float(min_value)
    max_value = float(max_value)
    prior = (max_value - min_value) * (prior - np.min(prior)) \
        / (np.max(prior) - np.min(prior)) + min_value
  # ====== normaize by ====== #
  if norm:
    prior = prior / np.sum(prior)
  # ====== set zero indices ====== #
  prior = prior.tolist()
  for i in zero_ids:
    prior.insert(i, 0)
  return np.array(prior)

# ===========================================================================
# Diagnose
# ===========================================================================
def _split_list(x, rng, train=0.6, idfunc=None, inc_test=True):
  # ====== shuffle input ====== #
  if idfunc is not None:
    x_id = defaultdict(list)
    for i in x:
      x_id[idfunc(i)].append(i)
  else:
    x_id = {i: [j] for i, j in enumerate(x)}
  # shuffle ID(s)
  id_list = list(x_id.keys())
  rng.shuffle(id_list)
  # ====== split ====== #
  N = len(id_list)
  if N == 1:
    raise ValueError("Only find 1 sample, cannot split")
  train = int(np.floor(float(train) * N))
  if train >= N:
    raise ValueError("train proportion must larger than 0 and smaller than 1.")
  valid = (N - train) // (2 if inc_test else 1)
  # ====== return splitted ====== #
  rets = (flatten_list((x_id[i] for i in id_list[:train]), level=1),
          flatten_list((x_id[i] for i in id_list[train: train + valid]), level=1))
  if inc_test:
    rets += (flatten_list((x_id[i] for i in id_list[train + valid:]), level=1),)
  else:
    rets += ([],)
  assert sum(len(r) for r in rets) == len(x), \
      "Number of returned data inconsitent from original data, %d != %d" % \
      (sum(len(r) for r in rets), len(x))
  return rets

def train_valid_test_split(x, train=0.6, cluster_func=None, idfunc=None,
                           inc_test=True, seed=None):
  """ Split given list into 3 dataset: for training, validating,
  and testing.

  Parameters
  ----------
  x: list, tuple, numpy.ndarray
      the dataset.
  train: float (0.0 - 1.0)
      proportion used for training, validating and testing will be
      half of the remain.
  cluster_func: None or call-able
      organize data into cluster, then applying the same
      train_valid_test split strategy for each cluster.
  idfunc: None or call-able
      a function transform the task list into unique identity for splitting,
      `idfunc` has to return comparable keys.
  inc_test: bool
      if split a proportion of data for testing also.
  seed: int
      random seed to produce a re-producible results.
  """
  rng = np.random.RandomState(seed)
  # ====== check input ====== #
  if isinstance(x, Mapping):
    x = x.items()
  elif isinstance(x, np.ndarray):
    x = x.tolist()
  # ====== check idfunc ====== #
  if not hasattr(idfunc, '__call__'):
    idfunc = None
  # ====== clustering ====== #
  if cluster_func is None:
    cluster_func = lambda x: 8 # lucky number
  if not hasattr(cluster_func, '__call__'):
    raise ValueError("'cluster_func' must be call-able or None.")
  clusters = defaultdict(list)
  for i in x:
    clusters[cluster_func(i)].append(i)
  # ====== applying data split for each cluster separately ====== #
  train_list, valid_list, test_list = [], [], []
  for name, clus in clusters.items():
    _1, _2, _3 = _split_list(clus, rng, train=train, idfunc=idfunc,
                             inc_test=inc_test)
    train_list += _1
    valid_list += _2
    test_list += _3
  # ====== return the results ====== #
  if inc_test:
    return train_list, valid_list, test_list
  return train_list, valid_list

def freqcount(x, key=None, count=1, normalize=False, sort=False,
              pretty_return=False):
  """ x: list, iterable

  Parameters
  ----------
  key: call-able
      extract the key from each item in the list
  count: call-able, int
      extract the count from each item in the list
  normalize: bool
      if normalize, all the values are normalized from 0. to 1. (
      which sum up to 1. in total).
  sort: boolean
      if True, the list will be sorted in ascent order.
  pretty_return: boolean
      if True, return pretty formatted text.

  Return
  ------
  dict: x(obj) -> freq(int)
  if `pretty_return` is `True`, return pretty formatted string.
  """
  freq = defaultdict(int)
  if key is None:
    key = lambda x: x
  if count is None:
    count = 1
  if isinstance(count, Number):
    _ = int(count)
    count = lambda x: _
  for i in x:
    c = count(i)
    i = key(i)
    freq[i] += c
  # always return the same order
  s = float(sum(v for v in freq.values()))
  freq = OrderedDict([(k, freq[k] / s if normalize else freq[k])
                      for k in sorted(freq.keys())])
  # check sort
  if sort:
    freq = OrderedDict(sorted(freq.items(), key=lambda x: x[1]))
  # check pretty return
  if pretty_return:
    s = ''
    for name, value in freq.items():
      s += ' %s: %d\n' % (ctext(name, 'yellow'), value)
    return s
  return freq

# ===========================================================================
# Bayesian
# ===========================================================================
def KL_divergence(P, Q):
  """ KL(P||Q) = ∑_i • p_i • log(p_i/q_i)
  The smaller this number, the better P match Q distribution
  """
  if isinstance(P, Mapping) and isinstance(Q, Mapping):
    keys = sorted(P.keys())
    P = [P[k] for k in keys]
    Q = [Q[k] for k in keys]
  # ====== normalize to probability 0-1 ====== #
  P = np.array(P)
  P = P / np.sum(P, axis=-1)
  Q = np.array(Q)
  Q = Q / np.sum(Q, axis=-1)
  # ====== calcuate the KL-div ====== #
  D = 0
  for pi, qi in zip(P, Q):
    D += pi * np.log(pi / qi)
  return D

# ===========================================================================
# Sampler
# ===========================================================================
def sampling_iter(it, k, p=None, return_iter=True, seed=1234,
                  progress_bar=None):
  """ Reservoir sampling, randomly choosing a sample of k items from a
  list S containing n items, where n is either a very large or unknown number.
  Typically n is large enough that the list doesn't fit into main memory.

  Parameters
  ----------
  it : iteration
    any instance of of iteration (i.e. `hasattr` '__iter__' )
  k : int
    number of sample for sampling
  p : {None, float (0.-1.)} (default: None)
    if `p` is None, perform reservoir sampling that guarantee every
    elements of the iteration is equally selected.
    if `p` is scalar, perform decayed sampling with `p` is the start
    probability (recommended in case `k` does not fit in the memory)
  return_iter : bool (default: True)
    if True, return an iteration of results instead of extracted list
  seed : int (default: 1234)
    random seed for reproducibility
  progress_bar : {None, odin.utils.Progbar.ProgBar}
  """
  k = int(k); assert k > 0
  if p is not None:
    p = float(p); assert 0. < p < 1.
  assert hasattr(it, '__iter__')
  # ====== check progress bar ====== #
  if progress_bar is not None:
    from odin.utils import Progbar
    assert isinstance(progress_bar, Progbar),\
    '`progress_bar` must be instance of odin.utils.progbar.Progbar, but given: %s' \
    % str(type(progress_bar))
  # ====== reservoir sampling ====== #
  if p is None:
    random.seed(seed)
    ret = []
    for i, x in enumerate(it):
      if i < k:
        ret.append(x)
      else:
        # as the iteration move forward,
        # the chance of picking new sample decrease
        r = random.randint(0, i)
        if r < k:
          ret[r] = x
      # update progress bar
      if progress_bar is not None:
        progress_bar.add(x)
    return tuple(ret)

  # ====== simulating the probability decay ====== #
  def _sampling():
    n = 0
    prob = p
    ret = []
    # this is compromise of randomness for speed
    n_buffer = 12000
    rand = np.random.RandomState(seed=seed)
    buffered_random = rand.rand(n_buffer)
    buffered_index = rand.randint(0, k, size=n_buffer, dtype=int)
    for i, x in enumerate(it):
      r = buffered_random[i % n_buffer]
      # initialize the reservoir
      if len(ret) < k:
        ret.append(x)
      # sample selected
      elif r < prob:
        yield x
        n += 1
      # update the reservoir
      else:
        ret[buffered_index[i % n_buffer]] = x
      # check break condition
      if n >= k:
        break
      # update the probability
      prob = 0.8 * prob + 0.2 / (i + 1)
      # update progress bar
      if progress_bar is not None:
        progress_bar.add(x)
    # return the rest of the samples to have enough k sample
    for i in range(k - n):
      yield ret[i]
  return _sampling() if return_iter else list(_sampling)

# ===========================================================================
# Statistics
# ===========================================================================
def sparsity_percentage(x, batch_size=1234):
  n_zeros = 0
  n_total = np.prod(x.shape)
  for start, end in batching(batch_size=batch_size, n=x.shape[0],
                             seed=None):
    y = x[start:end]
    if hasattr(y, 'count_nonzero'):
      n_nonzeros = y.count_nonzero()
    else:
      n_nonzeros = np.count_nonzero(y)
    n_zeros += np.prod(y.shape) - n_nonzeros
  return n_zeros / n_total

def logVMR(x, axis=None, logged_values=False):
  """ Calculate the variance to mean ratio (VMR) in non-logspace
  (return answer in log-space)

  VMR (variance-to-mean ratio = index of dispersion) is equal to zero
  in the case of a constant random variable (not dispersed).

  It is equal to one in a Poisson distribution, higher than one in a
  negative binomial distribution (over dispersed) and between zero
  and one in a binomial distribution (under dispersed).

  Reference
  ---------
  https://www.quantshare.com/item-1029-index-of-dispersion-variance-to-mean-ratio-vmr

  """
  if logged_values:
    x = np.expm1(x)
  return np.log1p(np.var(x, axis=axis) / np.mean(x, axis=axis))

# ===========================================================================
# Diagnosis
# ===========================================================================
def classification_diagnose(X, y_true, y_pred,
                            num_samples=8, return_list=False, top_n=None,
                            seed=1234):
  """
  Return
  ------
  OrderedDict: (true_class, pred_class) -> [list of samples from `X`]
  sorted by most frequence to less frequence
  """
  # Mapping from classesID -> [Sample ID]
  # ====== check argument ====== #
  if y_true.ndim == 2:
    y_true = np.argmax(y_true, axis=-1)
  elif y_true.ndim != 1:
    raise ValueError("Only support 1-D or 2-D one-hot `y_true`, "
                     "given `y_true` with shape: %s" % str(y_true.shape))
  if y_pred.ndim == 2:
    y_pred = np.argmax(y_pred, axis=-1)
  elif y_pred.ndim != 1:
    raise ValueError("Only support 1-D or 2-D one-hot `y_pred`, "
                     "given `y_pred` with shape: %s" % str(y_pred.shape))
  assert len(y_true) == len(y_pred) == len(X), "Inconsistent number of samples"
  # ====== initialize ====== #
  rand = np.random.RandomState(seed)
  miss = defaultdict(list)
  # ====== sampling ====== #
  for idx, (true, pred) in enumerate(zip(y_true, y_pred)):
    if true != pred:
      miss[(true, pred)].append(idx)
  # sort by the most freq mistake classes
  outputs = OrderedDict()
  for (true, pred), samples in sorted(miss.items(),
                                      key=lambda x: len(x[-1]),
                                      reverse=True):
    rand.shuffle(samples)
    outputs[(true, pred)] = (X[samples[:num_samples]]
                             if isinstance(X, np.ndarray) else
                             [X[i] for i in samples])
  # select top frequence
  if top_n is not None and isinstance(top_n, Number):
    top_n = int(top_n)
    assert top_n >= 1
    outputs = OrderedDict(list(outputs.items())[:top_n])
  return list(outputs.items()) if return_list else outputs

def classification_report(y_pred, y_true, labels):
  """
  Parameters
  ----------
  pass

  Return
  ------
  Classification report in form of string
  """
  from sklearn.metrics import accuracy_score, classification_report, confusion_matrix
  # ====== validate labels ====== #
  labels = as_tuple(labels)
  target_names = [str(i) for i in labels]
  labels = list(range(0, len(labels)))
  # ====== create report ====== #
  s = ""
  s += "Accuracy: %f\n" % accuracy_score(y_true, y_pred, normalize=True)
  s += "Confusion matrix:\n"
  s += str(confusion_matrix(y_true, y_pred, labels=labels)) + '\n'
  s += "Report:\n"
  s += str(classification_report(y_true, y_pred, labels=labels, digits=3,
                                 target_names=target_names))
  return s

def summary(x, axis=None, shorten=False, float_precision=2):
  """ Return string of statistical summary given series `x`
    {#:%s|mi:%s|q1:%s|md:%s|mn:%s|q3:%s|ma:%s|sd:%s}
  """
  if isinstance(x, Iterator):
    x = list(x)
  if isinstance(x, (tuple, list, set)):
    x = np.array(x)
  mean, std = np.mean(x, axis=axis), np.std(x, axis=axis)
  median = np.median(x, axis=axis)
  qu1, qu3 = np.percentile(x, [25, 75], axis=axis)
  min_, max_ = np.min(x, axis=axis), np.max(x, axis=axis)
  s = ""
  fmt = '%.' + str(int(float_precision)) + 'f'
  if not shorten:
    x = x.ravel()
    samples = ', '.join([str(i)
           for i in np.random.choice(x, size=min(8, len(x)), replace=False).tolist()])
    s += "***** Summary *****\n"
    s += "    Min : %s\n" % (fmt % min_)
    s += "1st Qu. : %s\n" % (fmt % qu1)
    s += " Median : %s\n" % (fmt % median)
    s += "   Mean : %s\n" % (fmt % mean)
    s += "3rd Qu. : %s\n" % (fmt % qu3)
    s += "    Max : %s\n" % (fmt % max_)
    s += "-------------------\n"
    s += "    Std : %s\n" % (fmt % std)
    s += "#Samples: %d\n" % len(x)
    s += "Samples : %s\n" % samples
    s += "Sparsity: %s\n" % (fmt % sparsity_percentage(x))
  else:
    s += "{#:%s|mi:%s|q1:%s|md:%s|mn:%s|q3:%s|ma:%s|sd:%s}" %\
    (ctext(len(x), 'cyan'),
     ctext(fmt % min_, 'cyan'),
     ctext(fmt % qu1, 'cyan'),
     ctext(fmt % median, 'cyan'),
     ctext(fmt % mean, 'cyan'),
     ctext(fmt % qu3, 'cyan'),
     ctext(fmt % max_, 'cyan'),
     ctext(fmt % std, 'cyan'))
  return s

describe = summary



# -*- coding: utf-8 -*-
# ===========================================================================
# This module is adpated from: https://github.com/glamp/bashplotlib
# Original work Copyright (c) 2013 Greg Lamp
# Modified work Copyright 2016-2017 TrungNT
# NOTE: the module is intentionally made for self-efficient, hence,
#       no need for external libraries
# ===========================================================================

from __future__ import print_function, absolute_import, division

import re
import math
import warnings
from collections import Mapping

import numpy as np

__all__ = [
    'remove_text_color',
    'escape_text_color',
    'merge_text_graph',
    'print_dist',
    'print_confusion',
    'print_hist',
    'print_bar',
    'print_scatter',
    'print_hinton'
]
# ===========================================================================
# Helper
# ===========================================================================
_chars = [" ", "▁", "▂", "▃", "▄", "▅", "▆", "▇", "█"]

isiterable = lambda x: hasattr(x, '__iter__') or hasattr(x, '__getitem__')

bcolours = {
    "white": '\033[97m',
    "aqua": '\033[96m',
    "pink": '\033[95m',
    "blue": '\033[94m',
    "yellow": '\033[93m',
    "green": '\033[92m',
    "red": '\033[91m',
    "grey": '\033[90m',
    "black": '\033[30m',
    "default": '\033[39m',
    "ENDC": '\033[39m',
}

colour_help = ', '.join([colour for colour in bcolours if colour != "ENDC"])


def get_colour(colour):
  """
  Get the escape code sequence for a colour
  """
  return bcolours.get(colour, bcolours['ENDC'])

def print_return_str(text, end='\n'):
  # if not return_str:
        # print(text, end=end)
  return text + end

def printcolour(text, sameline=False, colour=get_colour("ENDC")):
  """
  Print color text using escape codes
  """
  if sameline:
    sep = ''
  else:
    sep = '\n'
  if colour == 'default' or colour == 'ENDC' or colour is None:
    return print_return_str(text, sep)
  return print_return_str(get_colour(colour) + text + bcolours["ENDC"], sep)

def ctext(s, color='red'):
  try:
    from colorama import Fore
    color = color.upper()
    color = getattr(Fore, color, '')
    return color + s + Fore.RESET
  except ImportError:
    pass
  return s

def remove_text_color(s):
  s = re.sub(pattern=r"\\033\[\d\dm", repl='', string=s)
  return s

def escape_text_color(s):
  return remove_text_color(s)

def merge_text_graph(*graphs, padding=' '):
  """ To merge multiple graph together, this function
  will remove all the color coded text to properly align
  all the graph text.
  """
  padding = str(padding)
  assert len(graphs) >= 1
  if len(graphs) == 1:
    return graphs[0]

  def normalizing_lines(text):
    lines = [remove_text_color(l) for l in text.split('\n')]
    maxlen = max([len(l) for l in lines])
    lines = [l + ' ' * (maxlen - len(l))
             if len(l) < maxlen else l
             for l in lines]
    return lines

  graphs = [normalizing_lines(g) for g in graphs]
  maxlen = max(len(i) for i in graphs)
  final_text = ''
  for l in range(maxlen):
    for i, g in enumerate(graphs):
      if l < len(g):
        final_text += (padding if i > 0 else '') + g[l]
    final_text += '\n'
  return final_text

def drange(start, stop, step=1.0, include_stop=False):
  """
  Generate between 2 numbers w/ optional step, optionally include upper bound
  """
  if step == 0:
    step = 0.01
  r = start

  if include_stop:
    while r <= stop:
      yield r
      r += step
      r = round(r, 10)
  else:
    while r < stop:
      yield r
      r += step
      r = round(r, 10)


def box_text(text, width, offset=0):
  """
  Return text inside an ascii textbox
  """
  box = " " * offset + "-" * (width + 2) + "\n"
  box += " " * offset + "|" + text.center(width) + "|" + "\n"
  box += " " * offset + "-" * (width + 2)
  return box


def get_scale(series, is_y=False, steps=20):
  min_val = min(series)
  max_val = max(series)
  scaled_series = []
  for x in drange(min_val, max_val, (max_val - min_val) / steps,
                  include_stop=True):
    if x > 0 and scaled_series and max(scaled_series) < 0:
      scaled_series.append(0.0)
    scaled_series.append(x)

  if is_y:
    scaled_series.reverse()
  return scaled_series

def calc_bins(n, min_val, max_val, h=None, binwidth=None):
  """
  Calculate number of bins for the histogram
  """
  if not h:
    h = max(10, math.log(n + 1, 2))
  if binwidth == 0:
    binwidth = 0.1
  if binwidth is None:
    binwidth = (max_val - min_val) / h
  for b in drange(min_val, max_val, step=binwidth, include_stop=True):
    if b.is_integer():
      yield int(b)
    else:
      yield b

def read_numbers(numbers):
  """
  Read the input data in the most optimal way
  """
  if isiterable(numbers):
    for number in numbers:
      yield float(str(number).strip())
  else:
    for number in open(numbers):
      yield float(number.strip())

# ===========================================================================
# Main
# ===========================================================================
def print_dist(d, height=12, pch="o", show_number=False,
               title=None):
  """ Printing a figure of given distribution

  Parameters
  ----------
  d: dict, list
      a dictionary or a list, contains pairs of: "key" -> "count_value"
  height: int
      number of maximum lines for the graph
  pch : str
      shape of the bars in the plot, e.g 'o'

  Return
  ------
  str

  """
  LABEL_COLOR = ['cyan', 'yellow', 'blue', 'magenta', 'green']
  MAXIMUM_YLABEL = 4
  try:
    if isinstance(d, Mapping):
      d = d.items()
    orig_d = [(str(name), int(count))
              for name, count in d]
    d = [(str(name)[::-1].replace('-', '|').replace('_', '|'), count)
         for name, count in d]
    labels = [[c for c in name] for name, count in d]
    max_labels = max(len(name) for name, count in d)
    max_count = max(count for name, count in d)
    min_count = min(count for name, count in d)
  except Exception as e:
    raise ValueError('`d` must be distribution dictionary contains pair of: '
                     'label_name -> disitribution_count, error: "%s"' % str(e))
  # ====== create figure ====== #
  # draw height, 1 line for minimum bar, 1 line for padding the label,
  # then the labels
  nb_lines = int(height) + 1 + 1 + max_labels
  unit = (max_count - min_count) / height
  fig = ""
  # ====== add unit and total ====== #
  fig += ctext("Unit: ", 'red') + \
      '10^%d' % max(len(str(max_count)) - MAXIMUM_YLABEL, 0) + '  '
  fig += ctext("Total: ", 'red') + \
      str(sum(count for name, count in d)) + '\n'
  # ====== add the figure ====== #
  for line in range(nb_lines):
    value = max_count - unit * line
    # draw the y_label
    if line % 2 == 0 and line <= int(height): # value
      fig += ctext(
          ('%' + str(MAXIMUM_YLABEL) + 's') % str(int(value))[:MAXIMUM_YLABEL],
          color='red')
    else: # blank
      fig += ' ' * MAXIMUM_YLABEL
    fig += '|' if line <= int(height) else ' '
    # draw default line
    if line == int(height):
      fig += ''.join([ctext(pch + ' ',
                            color=LABEL_COLOR[i % len(LABEL_COLOR)])
                      for i in range(len(d))])
    # draw seperator for the label
    elif line == int(height) + 1:
      fig += '-' * (len(d) * 2)
    # draw the labels
    elif line > int(height) + 1:
      for i, lab in enumerate(labels):
        fig += ctext(' ' if len(lab) == 0 else lab.pop(),
                     LABEL_COLOR[i % len(LABEL_COLOR)]) + ' '
    # draw the histogram
    else:
      for i, (name, count) in enumerate(d):
        fig += ctext(pch if count - value >= 0 else ' ',
                     LABEL_COLOR[i % len(LABEL_COLOR)]) + ' '
    # new line
    fig += '\n'
  # ====== add actual number of necessary ====== #
  maximum_fig_length = MAXIMUM_YLABEL + 1 + len(orig_d) * 2
  if show_number:
    line_length = 0
    name_fmt = '%' + str(max_labels) + 's'
    for name, count in orig_d:
      n = len(name) + len(str(count)) + 4
      text = ctext(name_fmt % name, 'red') + ': %d ' % count
      if line_length + n >= maximum_fig_length:
        fig += '\n'
        line_length = n
      else:
        line_length += n
      fig += text
  # ====== add title ====== #
  if title is not None:
    title = ctext('"%s"' % str(title), 'red')
    padding = '  '
    n = (maximum_fig_length - len(title) // 2) // 2 - len(padding) * 3
    fig = '=' * n + padding + title + padding + '=' * n + '\n' + fig
  return fig[:-1]

def _float2str(x, fp=2):
  fmt = '%.' + '%d' % fp + 'f'
  return '1.' + '0' * (fp - 1) \
  if x > 0.99 else (fmt % x)[1:]

def print_confusion(arr, labels=None, side_bar=True, inc_stats=True,
                    float_precision=2):
  """
  Parameters
  ----------
  side_bar : bool (default: True)
    if True, include the side bar of precision, recall, F1,
    false alarm and number of samples

  inc_stats : bool (default: True)
    if True, include Precision, Recall, F1 ,False Alarm

  float_precision : int (.. > 0)
    number of floating point precision for printing the entries
  """
  side_bar = bool(side_bar)
  inc_stats = bool(inc_stats)
  float_precision = int(float_precision)
  assert float_precision > 0
  # ====== preprocessing ====== #
  LABEL_COLOR = 'magenta'
  if arr.ndim != 2:
    raise ValueError("Can only process 2-D confusion matrixf")
  if arr.shape[0] != arr.shape[1]:
    raise ValueError("`arr` must be square matrix.")
  nb_classes = arr.shape[0]
  if labels is None:
    labels = ['%d' % i for i in range(nb_classes)]
  else:
    labels = [str(i) for i in labels]
  max_label_length = max(len(i) for i in labels)
  lab_fmt = '%-' + str(max_label_length) + 's '
  # ====== calculate precision, recall, F1 ====== #
  arr_sum_row = arr.sum(-1).astype('float64')
  total_samples = np.sum(arr_sum_row)
  arr_sum_col = arr.sum(0).astype('float64')
  info = {}
  nb_info = 5 # Precision, Recall, F1, FA, Sum
  if inc_stats or side_bar:
    for i in range(nb_classes):
      TP = arr[i, i] # True positive
      FN = arr_sum_row[i] - arr[i, i] # False negative
      FP = arr_sum_col[i] - arr[i, i] # False positive
      precision = 0. if (TP + FP) == 0 else TP / (TP + FP)
      recall = 0. if (TP + FN) == 0 else TP / (TP + FN)
      f1 = 0. if precision == 0. or recall == 0. else \
          2 / (1 / precision + 1 / recall)
      fa = 0. if (total_samples - arr_sum_row[i]) == 0 else \
          FP / (total_samples - arr_sum_row[i]) # False alarm
      info[i] = (precision, recall, f1, fa, arr_sum_row[i])
  # normalize the confusion
  with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    arr = np.nan_to_num(arr.astype('float64') / arr_sum_row[:, None])
  # print title of the side bar
  fig = " " * ((3 + 1) * nb_classes + max_label_length + 2)
  if side_bar:
    fig += ctext('|'.join(['Pre', 'Rec', ' F1', ' FA', 'Sum']), 'red') + '\n'
    longest_line = len(fig) - 1
  else:
    longest_line = len(fig) - 1
    fig = ""
  # confusion matrix
  for row_id, row in enumerate(arr):
    row_text = ctext(lab_fmt % labels[row_id], LABEL_COLOR)
    # get the worst performed point
    most_misclassified = np.argsort(row, kind='quicksort').tolist()
    most_misclassified.remove(row_id)
    most_misclassified = most_misclassified[-1]
    if row[most_misclassified] == 0.:
      most_misclassified = None
    # iterate over each column value
    for col_id, col in enumerate(row):
      col = _float2str(col, fp=float_precision)
      if col_id == row_id:
        row_text += ctext(col,
                          color='blue' if col == '1.0' else 'cyan') + ' '
      elif most_misclassified is not None and \
      col_id == most_misclassified:
        row_text += ctext(col, color='red') + ' '
      else:
        row_text += col + ' '
    # new line
    if side_bar:
      # print float, except the last one is int
      info_str = [_float2str(val) if i < (nb_info - 1) else
                  ('%d' % np.round(val))
                  for i, val in enumerate(info[row_id])]
      fig += row_text + ' ' + '|'.join(info_str) + '\n'
    else:
      fig += row_text + '\n'
  # ====== draw labels at the bottom ====== #
  labels = [[c for c in i.replace('-', '|').replace('_', '|')[::-1]]
            for i in labels]
  for i in range(max_label_length):
    fig += ' ' * (max_label_length + 1)
    row = ''
    for l in labels:
      row += ' ' + (l.pop() if len(l) > 0 else ' ') + ' ' * float_precision
    fig += ctext(row, 'magenta') + '\n'
  # ====== Add the average values of stats ====== #
  if inc_stats:
    n = 0
    for i, name in enumerate(['Pre', 'Rec', ' F1', ' FA', '  Sum']):
      avr = np.mean([stats[i] for stats in info.values()])
      avr = '%.4g' % avr
      if avr[0] == '0':
        avr = avr[1:]
      avr = ctext('%s:' % name, 'red') + avr
      if n + len(avr) >= longest_line: # new line
        avr += '\n'
        n = 0
      else: # same line
        avr += '\t'
      n += len(avr) + len('\t')
      fig += avr
  return fig[:-1]

def print_hist(f, height=20.0, bincount=None, binwidth=None, pch="o",
    colour="default", title="", xlab=True, showSummary=False,
    regular=False):
  ''' Plot histogram.
   1801|       oo
   1681|       oo
   1561|      oooo
    961|      oooo
    841|      oooo
    721|     ooooo
    601|     oooooo
    241|     oooooo
    121|    oooooooo
      1| oooooooooooooo
        --------------
  Parameters
  ----------
  f : list(number), numpy.ndarray, str(filepath)
      input array
  height : float
      the height of the histogram in # of lines
  bincount : int
      number of bins in the histogram
  binwidth : int
      width of bins in the histogram
  pch : str
      shape of the bars in the plot, e.g 'o'
  colour : str
      white,aqua,pink,blue,yellow,green,red,grey,black,default,ENDC
  title : str
      title at the top of the plot, None = no title
  xlab : boolean
      whether or not to display x-axis labels
  showSummary : boolean
      whether or not to display a summary
  regular : boolean
      whether or not to start y-labels at 0
  return_str : boolean
      return string represent the plot or print it out, default: False
  '''
  if pch is None:
    pch = "o"
  splot = ''
  if isinstance(f, str):
    f = open(f).readlines()

  min_val, max_val = None, None
  n, mean, sd = 0.0, 0.0, 0.0

  for number in read_numbers(f):
    n += 1
    if min_val is None or number < min_val:
      min_val = number
    if max_val is None or number > max_val:
      max_val = number
    mean += number

  mean /= n

  for number in read_numbers(f):
    sd += (mean - number)**2

  sd /= (n - 1)
  sd **= 0.5

  bins = list(calc_bins(n, min_val, max_val, bincount, binwidth))
  hist = dict((i, 0) for i in range(len(bins)))

  for number in read_numbers(f):
    for i, b in enumerate(bins):
      if number <= b:
        hist[i] += 1
        break
    if number == max_val and max_val > bins[len(bins) - 1]:
      hist[len(hist) - 1] += 1

  min_y, max_y = min(hist.values()), max(hist.values())

  start = max(min_y, 1)
  stop = max_y + 1

  if regular:
    start = 1

  if height is None:
    height = stop - start
    if height > 20:
      height = 20

  ys = list(drange(start, stop, float(stop - start) / height))
  ys.reverse()

  nlen = max(len(str(min_y)), len(str(max_y))) + 1

  if title:
    splot += print_return_str(box_text(title, max(len(hist) * 2, len(title)), nlen))
  splot += print_return_str('')

  used_labs = set()
  for y in ys:
    ylab = str(int(y))
    if ylab in used_labs:
      ylab = ""
    else:
      used_labs.add(ylab)
    ylab = " " * (nlen - len(ylab)) + ylab + "|"

    splot += print_return_str(ylab, end=' ')

    for i in range(len(hist)):
      if int(y) <= hist[i]:
        splot += printcolour(pch, True, colour)
      else:
        splot += printcolour(" ", True, colour)
    splot += print_return_str('')
  xs = hist.keys()

  splot += print_return_str(" " * (nlen + 1) + "-" * len(xs))

  if xlab:
    xlen = len(str(float((max_y) / height) + max_y))
    for i in range(0, xlen):
      splot += printcolour(" " * (nlen + 1), True, colour)
      for x in range(0, len(hist)):
        num = str(bins[x])
        if x % 2 != 0:
          pass
        elif i < len(num):
          splot += print_return_str(num[i], end=' ')
        else:
          splot += print_return_str(" ", end=' ')
      splot += print_return_str('')

  center = max(map(len, map(str, [n, min_val, mean, max_val])))
  center += 15

  if showSummary:
    splot += print_return_str('')
    splot += print_return_str("-" * (2 + center))
    splot += print_return_str("|" + "Summary".center(center) + "|")
    splot += print_return_str("-" * (2 + center))
    summary = "|" + ("observations: %d" % n).center(center) + "|\n"
    summary += "|" + ("min value: %f" % np.min(f)).center(center) + "|\n"
    summary += "|" + ("mean : %f" % np.mean(f)).center(center) + "|\n"
    summary += "|" + (" 1-per: %f" % np.percentile(f, 1)).center(center) + "|\n"
    summary += "|" + (" 5-per: %f" % np.percentile(f, 5)).center(center) + "|\n"
    summary += "|" + ("25-per: %f" % np.percentile(f, 25)).center(center) + "|\n"
    summary += "|" + ("median: %f" % np.median(f)).center(center) + "|\n"
    summary += "|" + ("75-per: %f" % np.percentile(f, 75)).center(center) + "|\n"
    summary += "|" + ("95-per: %f" % np.percentile(f, 95)).center(center) + "|\n"
    summary += "|" + ("99-per: %f" % np.percentile(f, 99)).center(center) + "|\n"
    summary += "|" + ("sd : %f" % np.std(f)).center(center) + "|\n"
    summary += "|" + ("max value: %f" % np.max(f)).center(center) + "|\n"
    summary += "-" * (2 + center)
    splot += print_return_str(summary)
  return splot

def print_bar(f, height=20.0, bincount=None, binwidth=None, pch="o",
    colour="default", title="", xlab=None, showSummary=False,
    regular=False):
  ''' Plot bar.

  Parameters
  ----------
  f : list(number), numpy.ndarray, str(filepath)
      input array
  height : float
      the height of the histogram in # of lines
  bincount : int
      number of bins in the histogram
  binwidth : int
      width of bins in the histogram
  pch : str
      shape of the bars in the plot, e.g 'o'
  colour : str
      white,aqua,pink,blue,yellow,green,red,grey,black,default,ENDC
  title : str
      title at the top of the plot, None = no title
  xlab : boolean
      whether or not to display x-axis labels
  showSummary : boolean
      whether or not to display a summary
  regular : boolean
      whether or not to start y-labels at 0

  Example
  -------
  >>> y = np.random.rand(50)
  >>> bash_bar(y, bincount=50, colour='red')

  >>> 0.971|
  >>> 0.923|                o         o             o   o
  >>> 0.875|                o   o     o           o o  oo
  >>> 0.827|                o   o     o     o     o o ooo
  >>> 0.779|   o            o   o     o   o o     o o ooo
  >>> 0.731|   o    o       o   o     o   o o     o o ooo
  >>> 0.683|   oo   oo      o   o o o o   o o     o o ooo
  >>> 0.635|   oo   oo      o  oo o o o   o o    oo o ooo
  >>> 0.587|   oo   oo      o  oo o ooo   o o    oo o ooo
  >>> 0.539|   oo   oo      o  oooo ooo   o o    oo o ooo o
  >>> 0.491|   ooo ooo      o  oooo ooo   o o   ooo o ooo o
  >>> 0.443|   ooo ooo     oo ooooo oooo  o o   ooo o ooo o
  >>> 0.395|   ooo ooo    ooooooooo oooo  o o   ooo o ooo o
  >>> 0.347|  oooo ooo    ooooooooo ooooo o o   ooo o ooo o
  >>> 0.299|  oooo ooo    ooooooooo ooooo o o   ooo o ooo o
  >>> 0.251|  oooo ooo    ooooooooo ooooo o o   ooo o ooo o o
  >>> 0.203|  oooo ooo    ooooooooo ooooo ooo   ooo o ooo o o
  >>> 0.155|  oooo ooo oo ooooooooo ooooo ooo   ooo ooooooo oo
  >>> 0.107| ooooo ooo oo ooooooooo oooooooooo  ooooooooooo ooo
  >>> 0.059| ooooo oooooo ooooooooo ooooooooooooooooooooooo ooo
  >>> 0.011| oooooooooooooooooooooo ooooooooooooooooooooooooooo
  >>>       --------------------------------------------------
  '''
  if len(f) == 1:
    f = [min(0., np.min(f))] + [i for i in f]

  if pch is None:
    pch = "o"

  splot = ''
  if isinstance(f, str):
    f = open(f).readlines()

  # ====== Create data ====== #
  min_val, max_val = None, None
  n, mean, sd = 0.0, 0.0, 0.0

  # pick mode and get data
  numbers = [i for i in read_numbers(f)]
  int_mode = False
  if numbers[0].is_integer():
    int_mode = True

  # rescale big enough to show on bars
  min_orig = min(numbers) # original
  max_orig = max(numbers)
  numbers = [1000 * (i - min_orig) / (max_orig - min_orig + 1e-8) for i in numbers]

  # statistics
  n = len(numbers)
  min_val = min(numbers)
  max_val = max(numbers)
  mean = sum(numbers) / n
  sd = (sum([(mean - i)**2 for i in numbers]) / (n - 1)) ** 0.5

  # bins is index
  if bincount is not None:
    bincount = min(bincount, n)
  bins = list(calc_bins(n, 0., n + 0., bincount, binwidth))
  bins = [int(i) for i in bins]
  hist = dict((i, 0) for i in range(len(bins) - 1))

  # hist is the mean value of array with indices within bin
  for idx, (i, j) in enumerate(zip(bins, bins[1:])):
    if i == j: j += 1
    arr = numbers[i:j]
    hist[idx] = sum(arr) / len(arr) # calculate mean

  # ====== Start plot ====== #
  min_y, max_y = min(hist.values()), max(hist.values())

  start = max(min_y, 1)
  stop = max_y + 1

  if regular:
    start = 1

  if height is None:
    height = stop - start
    if height > 20:
      height = 20

  ys = list(drange(start, stop, float(stop - start) / height))
  ys.reverse()

  nlen = max(len(str(min_y)), len(str(max_y))) + 1

  if title:
    splot += print_return_str(
        box_text(title, max(len(hist) * 2, len(title)), nlen))
  splot += print_return_str('')

  used_labs = set()
  for y in ys:
    if int_mode:
      ylab = '%d' % int(y * (max_orig - min_orig + 1e-8) / 1000 + min_orig)
    else:
      ylab = '%.3f' % float(y * (max_orig - min_orig + 1e-8) / 1000 + min_orig)
    if ylab in used_labs:
      ylab = ""
    else:
      used_labs.add(ylab)
    ylab = " " * (nlen - len(ylab)) + ylab + "|"

    splot += print_return_str(ylab, end=' ')

    for i in range(len(hist)):
      if int(y) <= hist[i]:
        splot += printcolour(pch, True, colour)
      else:
        splot += printcolour(" ", True, colour)
    splot += print_return_str('')
  xs = hist.keys()

  splot += print_return_str(" " * (nlen + 1) + "-" * len(xs))

  if xlab:
    xlen = len(str(float((max_y) / height) + max_y))
    for i in range(0, xlen):
      splot += printcolour(" " * (nlen + 1), True, colour)
      for x in range(0, len(hist)):
        num = str(bins[x])
        if x % 2 != 0:
          pass
        elif i < len(num):
          splot += print_return_str(num[i], end=' ')
        else:
          splot += print_return_str(" ", end=' ')
      splot += print_return_str('')

  center = max(map(len, map(str, [n, min_val, mean, max_val])))
  center += 15

  if showSummary:
    splot += print_return_str('')
    splot += print_return_str("-" * (2 + center))
    splot += print_return_str("|" + "Summary".center(center) + "|")
    splot += print_return_str("-" * (2 + center))
    summary = "|" + ("observations: %d" % n).center(center) + "|\n"
    summary += "|" + ("min value: %f" % np.min(f)).center(center) + "|\n"
    summary += "|" + ("mean : %f" % np.mean(f)).center(center) + "|\n"
    summary += "|" + ("sd : %f" % np.std(f)).center(center) + "|\n"
    summary += "|" + ("max value: %f" % np.max(f)).center(center) + "|\n"
    summary += "-" * (2 + center)
    splot += print_return_str(summary)

  return splot


def print_scatter(xs, ys, size=None, pch='o',
                colour='red', title=None):
  ''' Scatter plot.
  ----------------------
  |                 *   |
  |               *     |
  |             *       |
  |           *         |
  |         *           |
  |        *            |
  |       *             |
  |      *              |
  -----------------------
  Parameters
  ----------
  xs : list, numpy.ndarray
      list of x series
  ys : list, numpy.ndarray
      list of y series
  size : int
      width of plot
  pch : str
      any character to represent a points
  colour : str, list(str)
      white,aqua,pink,blue,yellow,green,red,grey,black,default,ENDC
  title : str
      title for the plot, None = not show
  return_str : boolean
      return string represent the plot or print it out, default: False
  '''
  splot = ''
  plotted = set()
  cs = colour

  if size is None:
    size = 13

  if title:
    splot += print_return_str(box_text(title, 2 * len(get_scale(xs, False, size)) + 1))

  # ====== Top line ====== #
  splot += print_return_str(' ' + "-" * (len(get_scale(xs, False, size)) + 2))
  # ====== Main plot ====== #
  for y in get_scale(ys, True, size):
    splot += print_return_str("|", end=' ')
    for x in get_scale(xs, False, size):
      point = " "
      for (i, (xp, yp)) in enumerate(zip(xs, ys)):
        if xp <= x and yp >= y and (xp, yp) not in plotted:
          point = pch
          plotted.add((xp, yp))
          if isinstance(cs, list):
            colour = cs[i]
      splot += printcolour(point, True, colour)
    splot += print_return_str(" |")
  # ====== Bottom line ====== #
  splot += print_return_str(' ' + "-" * (len(get_scale(xs, False, size)) + 2))
  return splot


def print_hinton(arr, max_arr=None):
  ''' Print bar string, fast way to visual magnitude of value in terminal

  Example:
  -------
  >>> W = np.random.rand(10,10)
  >>> print_hinton(W)
  >>> ▁▃▄█▅█ ▅▃▅
  >>> ▅▂▆▄▄ ▅▅
  >>> ▄██▆▇▆▆█▆▅
  >>> ▄▄▅▂▂▆▅▁▅▆
  >>> ▂ ▁  ▁▄▆▅▁
  >>> ██▃█▃▃▆ ▆█
  >>>  ▁▂▁ ▁▃▃▆▂
  >>> ▅▂▂█ ▂ █▄▅
  >>> ▃▆▁▄▁▆▇▃▅▁
  >>> ▄▁▇ ██▅ ▂▃
  Returns
  -------
  return : str
      plot of array, for example: ▄▅▆▇
  '''
  arr = np.asarray(arr)
  if len(arr.shape) == 1:
    arr = arr[None, :]

  def visual_func(val, max_val):
    if abs(val) == max_val:
      step = len(_chars) - 1
    else:
      step = int(abs(float(val) / max_val) * len(_chars))
    colourstart = ""
    colourend = ""
    if val < 0:
      colourstart, colourend = '\033[90m', '\033[0m'
    return colourstart + _chars[step] + colourend

  if max_arr is None:
    max_arr = arr
  max_val = max(abs(np.max(max_arr)), abs(np.min(max_arr)))
  # print(np.array2string(arr,
  #                       formatter={'float_kind': lambda x: visual(x, max_val)},
  #                       max_line_width=5000)
  # )
  f = np.vectorize(visual_func)
  result = f(arr, max_val) # array of ▄▅▆▇
  rval = ''
  for r in result:
    rval += ''.join(r) + '\n'

  return rval[:-1]



from __future__ import absolute_import, division, print_function

import tensorflow as tf
from tensorflow_probability.python.layers import (MixtureLogistic,
                                                  MixtureNormal,
                                                  MixtureSameFamily)

__all__ = [
    'MixtureLogisticLayer', 'MixtureNormalLayer', 'MixtureSameFamilyLayer'
]
MixtureLogisticLayer = MixtureLogistic
MixtureNormalLayer = MixtureNormal
MixtureSameFamilyLayer = MixtureSameFamily



from odin.bay import distribution_layers, distributions, mixed_membership
from odin.bay.distribution_alias import parse_distribution
# this is important utility
from odin.bay.distributions.utils import concat_distribution
from odin.bay.helpers import *



from __future__ import absolute_import, division, print_function

import collections

import tensorflow as tf
from tensorflow.python.keras.layers import Layer
from tensorflow_probability.python.distributions import Distribution
from tensorflow_probability.python.layers.distribution_layer import (
    DistributionLambda, _get_convert_to_tensor_fn, _serialize)
from tensorflow_probability.python.layers.internal import \
    distribution_tensor_coercible as dtc
from tensorflow_probability.python.layers.internal import \
    tensor_tuple as tensor_tuple

__all__ = [
    'ConcatDistribution', 'Sampling', 'Moments', 'Stddev', 'DistributionAttr'
]


def _check_distribution(x):
  assert isinstance(x, Distribution), \
  "Input to this layer must be instance of tensorflow_probability Distribution"


class ConcatDistribution(DistributionLambda):
  """ This layer create a new `Distribution` by concatenate parameters of
  multiple distributions of the same type along given `axis`
  """

  def __init__(self,
               axis=None,
               convert_to_tensor_fn=Distribution.sample,
               **kwargs):
    from odin.bay.distributions.utils import concat_distribution
    super(ConcatDistribution, self).__init__(
        lambda dists: concat_distribution(dists=dists, axis=axis),
        convert_to_tensor_fn, **kwargs)
    self.axis = axis

  def get_config(self):
    config = super(ConcatDistribution, self).get_config()
    config['axis'] = self.axis
    return config


class Sampling(Layer):
  """ Sample the output from tensorflow-probability
  distribution layers """

  def __init__(self, n_samples=None, **kwargs):
    super(Sampling, self).__init__(**kwargs)
    self.n_samples = n_samples

  def get_config(self):
    config = super(Sampling, self).get_config()
    config['n_samples'] = self.n_samples
    return config

  def call(self, x, n_samples=None, **kwargs):
    if not isinstance(x, Distribution):
      return tf.expand_dims(x, axis=0)

    if n_samples is None:
      n_samples = self.n_samples
    return x.sample() if n_samples is None else x.sample(n_samples)


class Moments(Layer):
  """ Moments """

  def __init__(self, mean=True, variance=True, **kwargs):
    super(Moments, self).__init__(**kwargs)
    self.mean = bool(mean)
    self.variance = bool(variance)
    assert self.mean or self.variance, "This layer must return mean or variance"

  def call(self, x, **kwargs):
    if not isinstance(x, Distribution):
      return x

    outputs = []
    if self.mean:
      outputs.append(x.mean())
    if self.variance:
      outputs.append(x.variance())
    return outputs[0] if len(outputs) == 1 else tuple(outputs)

  def get_config(self):
    config = super(Moments, self).get_config()
    config['mean'] = self.mean
    config['variance'] = self.variance
    return config

  def compute_output_shape(self, input_shape):
    return [input_shape, input_shape] \
      if self.mean and self.variance else input_shape


class Stddev(Layer):
  """ Get standard deviation of an input distribution, return identity
  if the input is not an instance of `Distribution` """

  def __init__(self, **kwargs):
    super(Stddev, self).__init__(**kwargs)

  def call(self, x, **kwargs):
    if not isinstance(x, Distribution):
      return x

    return x.stddev()

  def compute_output_shape(self, input_shape):
    return input_shape


class DistributionAttr(Layer):
  """ This layer provide convenient way to extract statistics stored
  as attributes of the `Distribution` """

  def __init__(self,
               attr_name,
               convert_to_tensor_fn=Distribution.sample,
               **kwargs):
    super(DistributionAttr, self).__init__(**kwargs)
    self.attr_name = str(attr_name)
    if isinstance(convert_to_tensor_fn, property):
      convert_to_tensor_fn = convert_to_tensor_fn.fget
    self.convert_to_tensor_fn = convert_to_tensor_fn

  def get_config(self):
    config = super(DistributionAttr, self).get_config()
    config['attr_name'] = self.attr_name
    config['convert_to_tensor_fn'] = self.convert_to_tensor_fn
    return config

  def call(self, x):
    attrs = self.attr_name.split('.')
    for a in attrs:
      x = getattr(x, a)
    # special case a distribution is returned
    if isinstance(x, Distribution) and not isinstance(x, dtc._TensorCoercible):
      dist = dtc._TensorCoercible(
          distribution=x, convert_to_tensor_fn=self.convert_to_tensor_fn)
      value = tf.convert_to_tensor(value=dist)
      value._tfp_distribution = dist
      dist.shape = value.shape
      dist.get_shape = value.get_shape
      x = dist
    return x



# -*- coding: utf-8 -*-
##################################################################
# Example of usage:
##################################################################
from __future__ import print_function, division, absolute_import

import sys
import time
import inspect
from numbers import Number
from datetime import datetime
from contextlib import contextmanager
from collections import OrderedDict, defaultdict

import numpy as np

from odin.visual.bashplot import print_bar, print_confusion

try:
  from tqdm import __version__ as tqdm_version
  tqdm_version = int(tqdm_version.split(".")[0])
  if tqdm_version < 4:
    raise ImportError
  from tqdm import tqdm as _tqdm
  from tqdm._utils import _environ_cols_wrapper
except ImportError:
  sys.stderr.write("[ERROR] Cannot import `tqdm` version >= 4.\n")
  exit()

try:
  import colorama
  colorama.init()
  from colorama import Fore as _Fore
  _RED = _Fore.RED
  _YELLOW = _Fore.YELLOW
  _CYAN = _Fore.CYAN
  _MAGENTA = _Fore.MAGENTA
  _RESET = _Fore.RESET
except ImportError:
  _RED, _YELLOW, _CYAN, _MAGENTA, _RESET = '', '', '', '', ''

_NUMBERS_CH = {
    ord('0'): 0,
    ord('1'): 1,
    ord('2'): 2,
    ord('3'): 3,
    ord('4'): 4,
    ord('5'): 5,
    ord('6'): 6,
    ord('7'): 7,
    ord('8'): 8,
    ord('9'): 9,
}

# ===========================================================================
# Helper
# ===========================================================================
_LAST_UPDATED_PROG = [None]

def add_notification(msg):
  msg = _CYAN + "[%s]Notification:" % \
      datetime.now().strftime('%d/%b-%H:%M:%S') + _RESET + msg + ''
  _tqdm.write(msg)

class _FuncWrap(object):

  def __init__(self, func, default_func=lambda x: x):
    super(_FuncWrap, self).__init__()
    if func is None:
      func = default_func
    assert inspect.isfunction(func), \
    "Invalid function object of type: %s" % str(type(func))
    self.func = func

  def __call__(self, *args, **kwargs):
    return self.func(*args, **kwargs)

  def __getstate__(self):
    import dill
    return dill.dumps(self.func)

  def __setstate__(self, states):
    import dill
    self.func = dill.loads(states)

def _default_dict_list_creator():
  return defaultdict(list)

# ===========================================================================
# Progress bar
# ===========================================================================
class Progbar(object):

  """ Comprehensive review of any progress, this object is
  fully pickle-able, and can be used for storing history,
  summaries and report of the progress as well.

  Parameters
  ----------
  target: int
      total number of steps expected

  interval: float
      Minimum progress display update interval, in seconds.

  keep: bool
      whether to keep the progress bar when the epoch finished

  print_report: bool
      print updated report along with the progress bar for each update

  print_summary: bool
      print epoch summary after each epoch

  count_func: call-able
      a function takes the returned batch and return an integer for upating
      progress.

  report_func: call-able
      a function takes the returned batch and a collection of pair
      (key, value) for constructing the report.

  progress_func : call-able
      for post-processing the return value during processing into
      a number representing addition in the progress

  name: str or None
      specific name for the progress bar

  Examples
  --------
  >>> import numpy as np
  >>> from odin.utils import Progbar
  >>> x = list(range(10))
  >>> for i in Progbar(target=x):
  ...     pass

  Note
  ----
  Some special case:
      * any report key contain "confusionmatrix" will be printed out using
      `print_confusion`
      * any report key
  """
  FP = sys.stderr

  def __init__(self, target, interval=0.08, keep=False,
               print_progress=True, print_report=True, print_summary=False,
               count_func=None, report_func=None, progress_func=None,
               name=None):
    self.__pb = None # tqdm object
    if isinstance(target, Number):
      self.target = int(target)
      self.__iter_obj = None
    elif hasattr(target, '__len__'):
      self.target = len(target)
      self.__iter_obj = target
    else:
      raise ValueError("Unsupport for `target` type: %s" %
                       str(target.__class__))

    self._seen_so_far = defaultdict(int) # mapping: epoch_idx -> seen_so_far

    n = len(str(self.target))
    self._counter_fmt = '(%%%dd/%%%dd)' % (n, n)

    if name is None:
      name = "Progress-%s" % datetime.utcnow()
    self._name = name
    # ====== flags ====== #
    self.__interval = float(interval)
    self.__keep = keep
    self.print_progress = bool(print_progress)
    self.print_report = bool(print_report)
    self.print_summary = bool(print_summary)
    # ====== for history ====== #
    self._report = OrderedDict()
    self._last_report = None
    self._last_print_time = None
    self._epoch_summarizer_func = {}
    # ====== recording history ====== #
    # dictonary: {epoch_id: {key: [value1, value2, ...]}}
    self._epoch_hist = defaultdict(_default_dict_list_creator)
    self._epoch_summary = defaultdict(dict)
    self._epoch_idx = 0
    self._epoch_start_time = None
    # ====== iter information ====== #
    if self.__iter_obj is None and \
    (count_func is not None or report_func is not None):
      raise RuntimeError("`count_func` and `report_func` can only be used "
                         "when `target` is an iterator with specific length.")
    #
    self.__count_func = _FuncWrap(func=count_func,
                                  default_func=lambda x: len(x))
    self.__report_func = _FuncWrap(func=report_func,
                                   default_func=lambda x: None)
    # ====== check progress function ====== #
    self._progress_func = _FuncWrap(func=progress_func,
                                    default_func=lambda x: x)
    # ====== other ====== #
    self._labels = None # labels for printing the confusion matrix

  # ==================== History management ==================== #
  def __getitem__(self, key):
    return self._report.__getitem__(key)

  def __setitem__(self, key, val):
    self._epoch_hist[self.epoch_idx][key].append(val)
    return self._report.__setitem__(key, val)

  def __delitem__(self, key):
    return self._report.__delitem__(key)

  def __iter__(self):
    if self.__iter_obj is None:
      raise RuntimeError("This Progbar cannot be iterated, "
                         "the set `target` must be iterable.")
    for X in self.__iter_obj:
      count = self.__count_func(X)
      report = self.__report_func(X)
      if report is not None:
        for key, val in report:
          self[key] = val
      self.add(int(count))
      yield X
    del self.__iter_obj
    del self.__count_func
    del self.__report_func

  # ==================== screen control ==================== #
  @property
  def epoch_idx(self):
    return self._epoch_idx

  @property
  def nb_epoch(self):
    return self._epoch_idx + 1

  @property
  def name(self):
    return self._name

  @property
  def labels(self):
    """ Special labels for printing the confusion matrix. """
    return self._labels

  @property
  def history(self):
    """ Return
    dictonary:
      {epoch_id : {tensor_name0: [batch_return1, batch_return2, ...],
                   tensor_name1: [batch_return1, batch_return2, ...],
                   ...},
       1 : {tensor_name0: [batch_return1, batch_return2, ...],
                  tensor_name1: [batch_return1, batch_return2, ...],
                  ...},
       ... }

    Example
    -------
    >>> for epoch_id, results in task.history.items():
    >>>   for tensor_name, values in results.items():
    >>>     print(tensor_name, len(values))
    """
    return self._epoch_hist

  def get_report(self, epoch=-1, key=None):
    if epoch < 0:
      epoch = self.nb_epoch + epoch - 1
    return self._epoch_hist[epoch] if key is None else \
    self._epoch_hist[epoch][key]

  def set_summarizer(self, key, fn):
    """ Epoch summarizer is a function, searching in the
    report for given key, and summarize all the stored values
    of each epoch into a readable format

    i.e. the input arguments is a list of stored epoch report,
    the output is a string.
    """
    if not hasattr(fn, '__call__'):
      raise ValueError('`fn` must be call-able.')
    key = str(key)
    self._epoch_summarizer_func[key] = _FuncWrap(func=fn, default_func=None)
    return self

  def set_name(self, name):
    self._name = str(name)
    return self

  def set_labels(self, labels):
    if labels is not None:
      self._labels = tuple([str(l) for l in labels])
    return self

  def _formatted_report(self, report_dict, margin='', inc_name=True):
    """ Convert a dictionary of key -> value to well formatted string."""
    if inc_name:
      text = _MAGENTA + "\t%s" % self.name + _RESET + '\n'
    else:
      text = ""
    report_dict = sorted(report_dict.items(), key=lambda x: str(x[0]))
    for i, (key, value) in enumerate(report_dict):
      # ====== check value of key and value ====== #
      key = margin + str(key).replace('\n', ' ')
      # ====== special cases ====== #
      if "confusionmatrix" in key.lower() or \
      "confusion_matrix" in key.lower() or \
      "confusion-matrix" in key.lower() or \
      "confusion matrix" in key.lower():
        value = print_confusion(value, labels=self.labels,
                                inc_stats=True)
      # just print out string representation
      else:
        value = str(value)
      # ====== multiple lines or not ====== #
      if '\n' in value:
        text += _YELLOW + key + _RESET + ":\n"
        for line in value.split('\n'):
          text += margin + ' ' + line + '\n'
      else:
        text += _YELLOW + key + _RESET + ": " + value + "\n"
    return text[:-1]

  @property
  def progress_bar(self):
    if self.__pb is None:
      it = range(self.target)
      self.__pb = _tqdm(iterable=it,
                    desc="Epoch%s" % str(self.epoch_idx),
                    leave=self.__keep, total=self.target,
                    file=Progbar.FP, unit='obj',
                    mininterval=self.__interval, maxinterval=10,
                    miniters=0, position=0)
      self.__pb.clear()
      self._epoch_start_time = time.time()
    return self.__pb

  @property
  def seen_so_far(self):
    return self._seen_so_far[self.epoch_idx]

  def _generate_epoch_summary(self, epoch, inc_name=False, inc_counter=True):
    seen_so_far = self._seen_so_far[epoch]
    if seen_so_far == 0:
      return ''
    # ====== include name ====== #
    if inc_name:
      s = _MAGENTA + "%s" % self.name + _RESET
    else:
      s = ""
    # ====== create epoch summary ====== #
    if seen_so_far == self.target: # epoch already finished
      speed = (1. / self._epoch_summary[epoch]['__avg_time__'])
      elapsed = self._epoch_summary[epoch]['__total_time__']
    else: # epoch hasn't finished
      avg_time = (time.time() - self._epoch_start_time) / self.seen_so_far \
      if self.progress_bar.avg_time is None else \
      self.progress_bar.avg_time
      speed = 1. / avg_time
      elapsed = time.time() - self._epoch_start_time
    # ====== counter ====== #
    if inc_counter:
      frac = seen_so_far / self.target
      counter_epoch = self._counter_fmt % (seen_so_far, self.target)
      percentage = "%6.2f%%%s " % (frac * 100, counter_epoch)
    else:
      percentage = ''
    s += _RED + " Epoch %d " % epoch + _RESET + "%.4f(s) %s%.4f(obj/s)" % \
        (elapsed, percentage, speed)
    # epoch summary
    summary = dict(self._epoch_summary[epoch])
    if len(summary) > 2:
      summary.pop('__total_time__', None)
      summary.pop('__avg_time__', None)
      s += '\n' + self._formatted_report(summary, margin='   ', inc_name=False)
    return s

  @property
  def summary(self):
    s = _MAGENTA + "Report \"%s\"    TotalEpoch: %d\n" % \
    (self.name, self.nb_epoch) + _RESET
    # ====== create summary for each epoch ====== #
    s += '\n'.join([self._generate_epoch_summary(i)
                    for i in range(self.nb_epoch)])
    return s[:-1]

  # ==================== same actions ==================== #
  def add_notification(self, msg):
    msg = _CYAN + "[%s][%s]Notification:" % \
        (datetime.now().strftime('%d/%b-%H:%M:%S'),
            _MAGENTA + self.name + _CYAN) + _RESET + msg
    _tqdm.write(msg)
    return self

  def _new_epoch(self):
    if self.__pb is None:
      return
    # calculate number of offset lines from last report
    if self._last_report is None:
      nlines = 0
    else:
      nlines = len(self._last_report.split('\n'))
    # ====== reset progress bar (tqdm) ====== #
    if self.__keep: # keep the last progress on screen
      self.__pb.moveto(nlines)
    else: # clear everything
      for i in range(nlines):
        Progbar.FP.write('\r')
        console_width = _environ_cols_wrapper()(Progbar.FP)
        Progbar.FP.write(' ' * (79 if console_width is None else console_width))
        Progbar.FP.write('\r')  # place cursor back at the beginning of line
        self.__pb.moveto(1)
      self.__pb.moveto(-(nlines * 2))
    self.__pb.close()
    # ====== create epoch summary ====== #
    for key, values in self._epoch_hist[self._epoch_idx].items():
      values = [v for v in values]
      # provided summarizer function
      if key in self._epoch_summarizer_func:
        self._epoch_summary[self._epoch_idx][key] = self._epoch_summarizer_func[key](values)
      # very heuristic way to deal with sequence of numbers
      elif isinstance(values[0], Number):
        self._epoch_summary[self._epoch_idx][key] = np.mean(values)
      # numpy array
      elif isinstance(values[0], np.ndarray):
        self._epoch_summary[self._epoch_idx][key] = sum(v for v in values)
    # total epoch time
    total_time = time.time() - self._epoch_start_time
    self._epoch_summary[self._epoch_idx]['__total_time__'] = total_time
    # average time for 1 object
    avg_time = self.__pb.avg_time
    if avg_time is None:
      avg_time = total_time / self.target
    self._epoch_summary[self._epoch_idx]['__avg_time__'] = avg_time
    # reset all flags
    self.__pb = None
    self._last_report = None
    self._last_print_time = None
    self._epoch_start_time = None
    self._epoch_idx += 1
    return self

  @contextmanager
  def safe_progress(self):
    """ This context manager will automatically call `pause` if the
    progress unfinished, hence, it doesn't mesh up the screen. """
    yield None
    if 0 < self.seen_so_far < self.target:
      self.pause()

  def pause(self):
    """ Call `pause` if progress is running, hasn't finish, and
    you want to print something else on the scree.
    """
    # ====== clear the report ====== #
    if self._last_report is not None:
      nlines = len(self._last_report.split("\n"))
      self.__pb.moveto(-nlines)
      for i in range(nlines):
        Progbar.FP.write('\r')
        console_width = _environ_cols_wrapper()(Progbar.FP)
        Progbar.FP.write(' ' * (79 if console_width is None else console_width))
        Progbar.FP.write('\r')  # place cursor back at the beginning of line
        self.__pb.moveto(1)
    else:
      nlines = 0
    # ====== clear the bar ====== #
    if self.__pb is not None:
      self.__pb.clear()
      self.__pb.moveto(-nlines)
    # ====== reset the last report ====== #
    # because we already clean everything,
    # set _last_report=None prevent
    # further moveto(-nlines) in add()
    self._last_report = None
    return self

  def add(self, n=1):
    """ You need to call pause if """
    n = self._progress_func(n)
    if not isinstance(n, Number):
      raise RuntimeError(
          "Progress return an object, but not given `progress_func` for post-processing")
    if n <= 0:
      return self
    fp = Progbar.FP
    # ====== update information ====== #
    seen_so_far = min(self._seen_so_far[self.epoch_idx] + n, self.target)
    self._seen_so_far[self.epoch_idx] = seen_so_far
    # ====== check last updated progress, for automatically pause ====== #
    if _LAST_UPDATED_PROG[0] is None:
      _LAST_UPDATED_PROG[0] = self
    elif _LAST_UPDATED_PROG[0] != self:
      _LAST_UPDATED_PROG[0].pause()
      _LAST_UPDATED_PROG[0] = self
    # ====== show report ====== #
    if self.print_report:
      curr_time = time.time()
      # update the report
      if self._last_print_time is None or \
      time.time() - self._last_print_time > self.__interval or\
      seen_so_far >= self.target:
        self._last_print_time = curr_time
        # move the cursor to last point
        if self._last_report is not None:
          nlines = len(self._last_report.split('\n'))
          self.progress_bar.moveto(-nlines)
        report = self._formatted_report(self._report)
        # clear old report
        if self._last_report is not None:
          for i, l in enumerate(self._last_report.split('\n')):
            fp.write('\r')
            fp.write(' ' * len(l))
            fp.write('\r')  # place cursor back at the beginning of line
            self.progress_bar.moveto(1)
          self.progress_bar.clear()
          self.progress_bar.moveto(-i - 1)
        fp.write(report)
        fp.flush()
        self._last_report = report
        self.progress_bar.moveto(1)
    # ====== show progress ====== #
    if self.print_progress:
      self.progress_bar.update(n=n)
    else:
      self.progress_bar
    # ====== end of epoch ====== #
    if seen_so_far >= self.target:
      self._new_epoch()
      # print summary of epoch
      if self.print_summary:
        _tqdm.write(self._generate_epoch_summary(self.epoch_idx - 1,
                                                 inc_name=True,
                                                 inc_counter=False))
    return self



from __future__ import print_function, division, absolute_import

import os
import re
import inspect
from enum import Enum

from six import add_metaclass, string_types
from abc import ABCMeta, abstractmethod
from collections import Mapping

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.pipeline import Pipeline, make_pipeline as _make_pipeline

from odin.fuel import Dataset
from odin.preprocessing.signal import delta, mvn, stack_frames
from odin.utils import (get_all_files, is_string, as_tuple, is_pickleable,
                        ctext, flatten_list, dummy_formatter,
                        get_formatted_datetime)

class ExtractorSignal(object):
  """ ExtractorSignal """

  def __init__(self):
    super(ExtractorSignal, self).__init__()
    self._timestamp = get_formatted_datetime(only_number=False)
    self._extractor = None
    self._msg = ''
    self._action = 'ignore'
    self._last_input = {}

  @property
  def message(self):
    return self._msg

  @property
  def action(self):
    return self._action

  def set_message(self, extractor, msg, last_input):
    if self._extractor is not None:
      raise RuntimeError(
          "This signal has stored message, cannot set message twice.")
    assert isinstance(extractor, Extractor), \
    '`extractor` must be instance of odin.preprocessing.base.Extractor, ' +\
    'but given type: %s' % str(type(extractor))
    self._extractor = extractor
    self._msg = str(msg)
    self._last_input = last_input
    return self

  def set_action(self, action):
    action = str(action).lower()
    assert action in ('warn', 'error', 'ignore'), \
    "`action` can be one of the following values: 'warn', 'error', 'ignore'; " + \
    "but given: %s" % action
    self._action = action
    return self

  def __str__(self):
    if self._extractor is None:
      raise RuntimeError("The Signal has not been configured by the Extractor")
    s = '[%s]' % self._timestamp
    s += '%s' % ctext(self._extractor.__class__.__name__, 'cyan') + '\n'
    s += 'Error message: "%s"' % ctext(self._msg, 'yellow') + '\n'
    s += 'Action: "%s"' % ctext(self._action, 'yellow') + '\n'
    # last input
    s += 'Last input: \n'
    if isinstance(self._last_input, Mapping):
      for k, v in sorted(self._last_input.items(),
                         key=lambda x: x[0]):
        s += '  %s: %s\n' % (ctext(str(k), 'yellow'), dummy_formatter(v))
    else:
      s += '  Type: %s\n' % ctext(type(self._last_input), 'yellow')
      s += '  Object: %s\n' % ctext(str(self._last_input), 'yellow')
    # parameters
    s += 'Attributes: \n'
    s += '  ' + ctext('InputLayer', 'yellow') + ': ' + str(self._extractor.is_input_layer) + '\n'
    s += '  ' + ctext('RobustLevel', 'yellow') + ': ' + self._extractor.robust_level + '\n'
    s += '  ' + ctext('InputName', 'yellow') + ': ' + str(self._extractor.input_name) + '\n'
    s += '  ' + ctext('OutputName', 'yellow') + ': ' + str(self._extractor.output_name) + '\n'
    for name, param in self._extractor.get_params().items():
      if name not in ('_input_name', '_output_name'):
        s += '  ' + ctext(name, 'yellow') + ': ' + dummy_formatter(param) + '\n'
    return s

# ===========================================================================
# Helper
# ===========================================================================
def make_pipeline(steps, debug=False):
  """ NOTE: this method automatically revmove None entries

   - Flatten list or dictionary found in steps.
   - Remove any object that not is instance of `Extractor`

  during creation of `Pipeline`.
  """
  ID = [0]

  def item2step(x):
    if isinstance(x, (tuple, list)):
      if len(x) == 1 and isinstance(x[0], Extractor):
        x = x[0]
        ID[0] += 1
        return (x.__class__.__name__ + str(ID[0]), x)
      elif len(x) == 2:
        if is_string(x[0]) and isinstance(x[1], Extractor):
          return x
        elif is_string(x[1]) and isinstance(x[0], Extractor):
          return (x[1], x[0])
    elif isinstance(x, Extractor):
      ID[0] += 1
      return (x.__class__.__name__ + str(ID[0]), x)
    return None

  if isinstance(steps, Mapping):
    steps = steps.items()
  elif not isinstance(steps, (tuple, list)):
    steps = [steps]
  steps = [item2step(i) for i in steps]
  # remove None
  steps = [s for s in steps if s is not None]
  if len(steps) == 0:
    raise ValueError("No instance of odin.preprocessing.base.Extractor found in `steps`.")
  # ====== set debug mode ====== #
  set_extractor_debug([i[1] for i in steps],
                      debug=bool(debug))
  # ====== return pipeline ====== #
  ret = Pipeline(steps=steps)
  return ret

def set_extractor_debug(extractors, debug):
  # ====== prepare ====== #
  if isinstance(extractors, (tuple, list)):
    extractors = [i for i in flatten_list(extractors)
                  if isinstance(i, Extractor)]
  elif isinstance(extractors, Pipeline):
    extractors = [i[-1] for i in extractors.steps]
  elif isinstance(extractors, Mapping):
    extractors = [i[-1] for i in extractors.items()]
  else:
    raise ValueError("No support for `extractors` type: %s" % type(extractors))
  # ====== set the value ====== #
  for i in extractors:
    i._debug = bool(debug)
  return extractors

def _equal_inputs_outputs(x, y):
  try:
    if x != y:
      return False
  except Exception:
    pass
  return True

def _preprocess(x):
  if isinstance(x, np.str_):
    x = str(x)
  return x

# ===========================================================================
# Basic extractors
# ===========================================================================
@add_metaclass(ABCMeta)
class Extractor(BaseEstimator, TransformerMixin):
  """ Extractor

  The developer must override the `_transform` method:
   - If the return is instance of `collections.Mapping`, the new features
     will be merged into existed features dictionary (i.e. the input dictionary)
   - If the return is not instance of `Mapping`, the name of given classes will
     be used to name the returned features.
   - If `None` is returned, no `_transform` is called, just return None for
     the whole pipeline (i.e. None act as terminal signal)

  Parameters
  ----------
  input_name : {None, string, list of string}
    list of string represent the name of feature
  output_name : {None, string, list of string}
    default name for the output feature (in case the return is not
    instance of dictionary);
    If `input_name` is None and `output_name` is None, use lower
    case of class name as default
  is_input_layer : bool (default: False)
    An input layer accept any type of input to `transform`,
    otherwise, only accept a dictionary type as input.
  robust_level : {'ignore', 'warn', 'error'}
    'ignore' - ignore error files
    'warn' - warn about error file during processing
    'error' - raise Exception and stop processing
  """

  def __init__(self, input_name=None, output_name=None,
               is_input_layer=False, robust_level='ignore'):
    super(Extractor, self).__init__()
    self._debug = False
    self._is_input_layer = bool(is_input_layer)
    self._last_debugging_text = ''
    # ====== robust level ====== #
    robust_level = str(robust_level).lower()
    assert robust_level in ('ignore', 'warn', 'error'),\
    "`robust_level` can be one of the following values: " + \
    "'warn', 'error', 'ignore'; but given: %s" % robust_level
    self._robust_level = robust_level
    # ====== check input_name ====== #
    if input_name is None:
      pass
    elif isinstance(input_name, string_types):
      pass
    elif hasattr(input_name, '__iter__'):
      input_name = tuple([str(i).lower() for i in input_name])
    else:
      raise ValueError("No support for `input_name` type: %s" % str(type(input_name)))
    self._input_name = input_name
    # ====== check output_name ====== #
    if output_name is None:
      if input_name is None:
        output_name = self.__class__.__name__.lower()
      else:
        output_name = input_name
    elif isinstance(output_name, string_types):
      pass
    elif hasattr(output_name, '__iter__'):
      output_name = tuple([str(i).lower() for i in output_name])
    else:
      raise ValueError("No support for `output_name` type: %s" % str(type(output_name)))
    self._output_name = output_name

  @property
  def last_debugging_text(self):
    """ Return the last debugging information recorded during
    calling the `transform` method with `debug=True` """
    if not hasattr(self, '_last_debugging_text'):
      self._last_debugging_text = ''
    return self._last_debugging_text

  @property
  def input_name(self):
    return self._input_name

  @property
  def output_name(self):
    return self._output_name

  @property
  def is_input_layer(self):
    return self._is_input_layer

  @property
  def robust_level(self):
    return self._robust_level

  def set_debug(self, debug):
    self._debug = bool(debug)
    return self

  def fit(self, X, y=None):
    # Do nothing here
    return self

  @abstractmethod
  def _transform(self, X):
    raise NotImplementedError

  def transform(self, X):
    # NOTE: do not override this method
    if isinstance(X, ExtractorSignal):
      return X
    # ====== interpret different signal ====== #
    if X is None:
      return ExtractorSignal(
      ).set_message(extractor=self,
                    msg="`None` value is returned by extractor",
                    last_input=X
      ).set_action(self.robust_level)
    # ====== input layer ====== #
    if not self.is_input_layer and not isinstance(X, Mapping):
      err_msg = "the input to `Extractor.transform` must be instance of dictionary, " + \
      "but given type: %s" % str(type(X))
      return ExtractorSignal(
      ).set_message(extractor=self,
                    msg=err_msg,
                    last_input=X
      ).set_action(self.robust_level)
    # ====== the transformation ====== #
    if self.input_name is not None and isinstance(X, Mapping):
      for name in as_tuple(self.input_name, t=string_types):
        if name not in X:
          return ExtractorSignal(
          ).set_message(extractor=self,
                        msg="Cannot find features with name: %s" % name,
                        last_input=X
          ).set_action('error')
    y = self._transform(X)
    # if return Signal or None, no post-processing
    if isinstance(y, ExtractorSignal):
      return y
    if y is None:
      return ExtractorSignal(
      ).set_message(extractor=self,
                    msg="`None` value is returned by the extractor: %s" % self.__class__.__name__,
                    last_input=X
      ).set_action(self.robust_level)
    # ====== return type must always be a dictionary ====== #
    if not isinstance(y, Mapping):
      if isinstance(y, (tuple, list)):
        y = {i: j
             for i, j in zip(as_tuple(self.output_name, t=string_types),
                             y)}
      else:
        y = {self.output_name: y}
    # ====== Merge previous results ====== #
    # remove None values
    tmp = {}
    for name, feat in y.items():
      if any(c.isupper() for c in name):
        return ExtractorSignal(
        ).set_message(extractor=self,
                      msg="Name for features cannot contain upper case",
                      last_input=X
        ).set_action('error')
      if feat is None:
        continue
      tmp[name] = feat
    y = tmp
    # add old features extracted in X, but do NOT override new features in y
    if isinstance(X, Mapping):
      for name, feat in X.items():
        if any(c.isupper() for c in name):
          return ExtractorSignal(
          ).set_message(extractor=self,
                        msg="Name for features cannot contain upper case",
                        last_input=X
          ).set_action('error')
        if name not in y:
          y[name] = _preprocess(feat)
    # ====== print debug text ====== #
    # maybe someone implement __getstate__ and forget _debug
    if not hasattr(self, '_debug'):
      self._debug = False
    if self._debug:
      debug_text = ''
      debug_text += '%s %s\n' % (ctext("[Extractor]", 'cyan'),
                                 ctext(self.__class__.__name__, 'magenta'))
      # inputs
      if not _equal_inputs_outputs(X, y):
        debug_text += '  %s\n' % ctext("Inputs:", 'yellow')
        debug_text += '  %s\n' % ctext("-------", 'yellow')
        if isinstance(X, Mapping):
          for k, v in X.items():
            debug_text += '    %s : %s\n' % (ctext(k, 'blue'), dummy_formatter(v))
        else:
          debug_text += '    %s\n' % dummy_formatter(X)
      # outputs
      debug_text += '  %s\n' % ctext("Outputs:", 'yellow')
      debug_text += '  %s\n' % ctext("-------", 'yellow')
      if isinstance(y, Mapping):
        for k, v in y.items():
          debug_text += '    %s : %s\n' % (ctext(k, 'blue'), dummy_formatter(v))
      else:
        debug_text += '    %s\n' % dummy_formatter(y)
      # parameters
      for name, param in self.get_params().items():
        if name not in ('_input_name',
                        '_output_name'):
          debug_text += '  %s : %s\n' % (ctext(name, 'yellow'), dummy_formatter(param))
      self._last_debugging_text = debug_text
      print(debug_text)
    return y

# ===========================================================================
# General extractor
# ===========================================================================
class Converter(Extractor):

  """ Convert the value under `input_name` to a new value
  using `converter` function, and save the new value to
  the `output_name`.

  This could be mapping 1 -> 1 or many -> 1; in case of
  many to 1 mapping, the `converter` function will be
  call as `converter(*args)`

  Parameters
  ----------
  converter: {Mapping, call-able}
      convert `inputs['name'] = converter(inputs[keys])`

  """

  def __init__(self, converter, input_name='name', output_name='name'):
    super(Converter, self).__init__(input_name=as_tuple(input_name, t=string_types),
                                    output_name=str(output_name))
    # ====== check converter ====== #
    if not hasattr(converter, '__call__') and \
    not isinstance(converter, Mapping):
      raise ValueError("`converter` must be call-able.")
    # converter can be function or dictionary
    self.converter = converter

  def _transform(self, feat):
    X = [feat[name] for name in self.input_name]
    if hasattr(self.converter, '__call__'):
      name = self.converter(*X)
    else:
      name = self.converter[X[0] if len(X) == 1 else X]
    return {self.output_name: name}

class DeltaExtractor(Extractor):

  """ Extracting the delta coefficients given the axis

  Parameters
  ----------
  input_name : list of str
      list of all features name for calculating the delta
  width : int
      amount of frames taken into account for 1 delta
  order : list of int
      list of all delta order will be concatenate (NOTE: keep `0` in
      the list if you want to keep original features)
  axis : int
      which dimension calculating the delta
      (suggest time-dimension for acoustic features, i.e. axis=0)
  """

  def __init__(self, input_name, output_name=None,
               width=9, order=(0, 1), axis=0):
    super(DeltaExtractor, self).__init__(
        input_name=as_tuple(input_name, t=string_types),
        output_name=output_name)
    # ====== check width ====== #
    width = int(width)
    if width % 2 == 0 or width < 3:
      raise ValueError("`width` must be odd integer >= 3, give value: %d" % width)
    self.width = width
    # ====== check order ====== #
    self.order = as_tuple(order, t=int)
    # ====== axis ====== #
    self.axis = axis

  def _calc_deltas(self, X):
    all_deltas = delta(data=X, width=self.width, order=max(self.order), axis=self.axis)
    if not isinstance(all_deltas, (tuple, list)):
      all_deltas = (all_deltas,)
    else:
      all_deltas = tuple(all_deltas)
    all_deltas = (X,) + all_deltas
    all_deltas = tuple([d for i, d in enumerate(all_deltas)
                        if i in self.order])
    return np.concatenate(all_deltas, axis=-1)

  def _transform(self, feat):
    return [self._calc_deltas(feat[name])
            for name in self.input_name]

class EqualizeShape0(Extractor):
  """ EqualizeShape0
  The final length of all features is the `minimum length`.

  This Extractor shrink the shape of all given features in `feat_name`
  to the same length.

  Raise Error if given files is shorted than desire length

  Parameters
  ----------
  input_name: {None, list of string}
      list of features name will be used for calculating the
      running statistics.
      If None, calculate the statistics for all `numpy.ndarray`
  shrink_mode: 'center', 'left', 'right'
       center: remove data points from both left and right
       left: remove data points at the beginning (left)
       right: remove data points at the end (right)
  """

  def __init__(self, input_name=None, shrink_mode='right'):
    super(EqualizeShape0, self).__init__(
        input_name=as_tuple(input_name, t=string_types) if input_name is not None else None)
    shrink_mode = str(shrink_mode).lower()
    if shrink_mode not in ('center', 'left', 'right'):
      raise ValueError("shrink mode support include: center, left, right")
    self.shrink_mode = shrink_mode

  def _transform(self, feat):
    if self.input_name is None:
      X = []
      output_name = []
      for key, val in feat.items():
        if isinstance(val, np.ndarray) and val.ndim > 0:
          X.append(val)
          output_name.append(key)
    else:
      X = [feat[name] for name in self.input_name]
      output_name = self.input_name
    # ====== searching for desire length ====== #
    n = min(i.shape[0] for i in X)
    # ====== equalize ====== #
    equalized = {}
    for name, y in zip(output_name, X):
      # cut the features in left and right
      # if the shape[0] is longer
      if y.shape[0] != n:
        diff = y.shape[0] - n
        if diff < 0:
          print("Feature length: %d which is smaller "
                "than desire length: %d, feature name is '%s'" %
                (y.shape[0], n, X['name']))
          return None
        elif diff > 0:
          if self.shrink_mode == 'center':
            diff_left = diff // 2
            diff_right = diff - diff_left
            y = y[diff_left:-diff_right]
          elif self.shrink_mode == 'right':
            y = y[:-diff]
          elif self.shrink_mode == 'left':
            y = y[diff:]
      equalized[name] = y
    return equalized

class RunningStatistics(Extractor):
  """ Running statistics

  Parameters
  ----------
  input_name: {string, list of string}
      list of features name will be used for calculating the
      running statistics.
      If None, calculate the statistics for all `numpy.ndarray`
      with `ndim` > 0
  axis : int (default: 0)
      the axis for calculating the statistics
  prefix : ''
      the prefix append to 'sum1' and 'sum2'
  """

  def __init__(self, input_name=None, axis=0, prefix=''):
    super(RunningStatistics, self).__init__(
        input_name=as_tuple(input_name, t=string_types) if input_name is not None else None)
    self.axis = axis
    self.prefix = str(prefix)

  def get_sum1_name(self, feat_name):
    return '%s_%ssum1' % (feat_name, self.prefix)

  def get_sum2_name(self, feat_name):
    return '%s_%ssum2' % (feat_name, self.prefix)

  def _transform(self, feat):
    if self.input_name is None:
      X = []
      output_name = []
      for key, val in feat.items():
        if isinstance(val, np.ndarray) and val.ndim > 0:
          X.append(val)
          output_name.append(key)
    else:
      X = [feat[name] for name in self.input_name]
      output_name = self.input_name
    # ====== calculate the statistics ====== #
    for name, y in zip(output_name, X):
      # ====== SUM of x^1 ====== #
      sum1 = np.sum(y, axis=self.axis, dtype='float64')
      s1_name = self.get_sum1_name(name)
      if s1_name not in feat:
        feat[s1_name] = sum1
      else:
        feat[s1_name] += sum1
      # ====== SUM of x^2 ====== #
      sum2 = np.sum(np.power(y, 2), axis=self.axis, dtype='float64')
      s2_name = self.get_sum2_name(name)
      if s2_name not in feat:
        feat[s2_name] = sum2
      else:
        feat[s2_name] += sum2
    return feat

class AsType(Extractor):
  """ An extractor convert given features to given types

  Parameters
  ----------
  dtype : {string, numpy.dtype}
      desire type
  input_name: {string, list of string}
      list of features name will be used for calculating the
      running statistics.
      If None, calculate the statistics for all object `hasattr`
      'astype'
  exclude_pattern : {string, None}
      regular expression pattern to exclude all features with given
      name pattern, only used when `input_name=None`.
      By default, exclude all running statistics '*_sum1' and '*_sum2'
  """

  def __init__(self, dtype, input_name=None, exclude_pattern=".+\_sum[1|2]"):
    super(AsType, self).__init__(
        input_name=as_tuple(input_name, t=string_types) if input_name is not None else None)
    self.dtype = np.dtype(dtype)
    if isinstance(exclude_pattern, string_types):
      exclude_pattern = re.compile(exclude_pattern)
    else:
      exclude_pattern = None
    self.exclude_pattern = exclude_pattern

  def _transform(self, feat):
    # ====== preprocessing ====== #
    if self.input_name is None:
      X = []
      output_name = []
      for key, val in feat.items():
        if hasattr(val, 'astype'):
          if self.exclude_pattern is not None and \
          self.exclude_pattern.search(key):
            continue
          X.append(val)
          output_name.append(key)
    else:
      X = [feat[name] for name in self.input_name]
      output_name = self.input_name
    # ====== astype ====== #
    updates = {}
    for name, y in zip(output_name, X):
      updates[name] = y.astype(self.dtype)
    return updates

class DuplicateFeatures(Extractor):

  def __init__(self, input_name, output_name):
    super(DuplicateFeatures, self).__init__(
        input_name=as_tuple(input_name, t=string_types),
        output_name=as_tuple(output_name, t=string_types))

  def _transform(self, feat):
    return {out_name: feat[in_name]
            for in_name, out_name in zip(self.input_name, self.output_name)}

class RenameFeatures(Extractor):

  def __init__(self, input_name, output_name):
    super(RenameFeatures, self).__init__(
        input_name=as_tuple(input_name, t=string_types),
        output_name=as_tuple(output_name, t=string_types))

  def _transform(self, X):
    return X

  def transform(self, feat):
    if isinstance(feat, Mapping):
      for old_name, new_name in zip(self.input_name, self.output_name):
        if old_name in feat: # only remove if it exist
          X = feat[old_name]
          del feat[old_name]
          feat[new_name] = X
    return feat

class DeleteFeatures(Extractor):
  """ Remove features by name from extracted features dictionary """

  def __init__(self, input_name):
    super(DeleteFeatures, self).__init__()
    self._name = as_tuple(input_name, t=string_types)

  def _transform(self, X):
    return X

  def transform(self, feat):
    if isinstance(feat, Mapping):
      for name in self._name:
        if name in feat: # only remove if it exist
          del feat[name]
    return feat

# ===========================================================================
# Shape
# ===========================================================================
class StackFeatures(Extractor):
  """ Stack context (or splice multiple frames) into
  single vector.

  Parameters
  ----------
  n_context : int
      number of context frame on the left and right, the final
      number of stacked frame is `context * 2 + 1`
      NOTE: the stacking process, ignore `context` frames at the
      beginning on the left, and at the end on the right.
  input_name : {None, list of string}
      list of features name will be used for calculating the
      running statistics.
      If None, calculate the statistics for all `numpy.ndarray`
  mvn: bool
      if True, preform mean-variance normalization on input features.
  """

  def __init__(self, n_context, input_name=None):
    super(StackFeatures, self).__init__(
        input_name=as_tuple(input_name, t=string_types) if input_name is not None else None)
    self.n_context = int(n_context)
    assert self.n_context > 0

  def _transform(self, feat):
    if self.input_name is None:
      X = []
      output_name = []
      for key, val in feat.items():
        if isinstance(val, np.ndarray) and val.ndim > 0:
          X.append(val)
          output_name.append(key)
    else:
      X = [feat[name] for name in self.input_name]
      output_name = self.input_name
    # ====== this ====== #
    for name, y in zip(output_name, X):
      # stacking the context frames
      y = stack_frames(y, frame_length=self.n_context * 2 + 1,
                       step_length=1, keep_length=True,
                       make_contigous=True)
      feat[name] = y
    return feat



from __future__ import absolute_import, division, print_function

from tensorflow.python.keras import Sequential
from tensorflow.python.keras.layers import Activation, Dense, Layer
from tensorflow.python.util.tf_export import keras_export

__all__ = ['Identity', 'Parallel']


class Identity(Layer):

  def __init__(self, name=None):
    super(Identity, self).__init__(name=name)
    self.supports_masking = True

  def call(self, inputs, training=None):
    return inputs

  def compute_output_shape(self, input_shape):
    return input_shape


@keras_export('keras.models.Sequential', 'keras.Sequential')
class Parallel(Sequential):
  """ Similar design to keras `Sequential` but simultanously applying
  all the layer on the input and return all the results.

  This layer is important for implementing multitask learning.
  """

  def call(self, inputs, training=None, mask=None, **kwargs):  # pylint: disable=redefined-outer-name
    if self._is_graph_network:
      if not self.built:
        self._init_graph_network(self.inputs, self.outputs, name=self.name)
      return super(Parallel, self).call(inputs, training=training, mask=mask)

    outputs = []
    for layer in self.layers:
      # During each iteration, `inputs` are the inputs to `layer`, and `outputs`
      # are the outputs of `layer` applied to `inputs`. At the end of each
      # iteration `inputs` is set to `outputs` to prepare for the next layer.
      kw = {}
      argspec = self._layer_call_argspecs[layer].args
      if 'mask' in argspec:
        kw['mask'] = mask
      if 'training' in argspec:
        kw['training'] = training
      # support custom keyword argument also
      for k, v in kwargs.items():
        if k in argspec:
          kw[k] = v

      o = layer(inputs, **kw)
      outputs.append(o)

    return tuple(outputs)

  def compute_output_shape(self, input_shape):
    shape = []
    for layer in self.layers:
      shape.append(layer.compute_output_shape(input_shape))
    return tuple(shape)

  def compute_mask(self, inputs, mask):
    outputs = self.call(inputs, mask=mask)
    return [o._keras_mask for o in outputs]



"""Fairly basic set of tools for real-time data augmentation on image data.
Can easily be extended to include new transformations,
new preprocessing methods, etc...
LICENSE: https://github.com/fchollet/keras/blob/master/LICENSE
Original code: https://github.com/fchollet/keras/blob/master/keras/preprocessing/image.py
"""
from __future__ import print_function, division, absolute_import

import numpy as np
import scipy.ndimage as ndi

from PIL import Image

from odin.utils import as_tuple


def apply_transform(x,
                    transform_matrix,
                    fill_mode='nearest'):
  """Apply the image transformation specified by a matrix.

  # Arguments
      x: 2D numpy array, single image.
      transform_matrix: Numpy array specifying the geometric transformation.
      channel_axis: Index of axis for channels in the input tensor.
      fill_mode: Points outside the boundaries of the input
          are filled according to the given mode
          (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
      cval: Value used for points outside the boundaries
          of the input if `mode='constant'`.

  # Returns
      The transformed version of the input.
  """
  x = np.rollaxis(x, 2, 0)
  final_affine_matrix = transform_matrix[:2, :2]
  final_offset = transform_matrix[:2, 2]
  channel_images = [ndi.interpolation.affine_transform(
      x_channel,
      final_affine_matrix,
      final_offset,
      order=0,
      mode=fill_mode,
      cval=0.) for x_channel in x]
  x = np.stack(channel_images, axis=0)
  x = np.rollaxis(x, 0, 2 + 1)
  return x


def transform_matrix_offset_center(matrix, x, y):
  o_x = float(x) / 2 + 0.5
  o_y = float(y) / 2 + 0.5
  offset_matrix = np.array([[1, 0, o_x], [0, 1, o_y], [0, 0, 1]])
  reset_matrix = np.array([[1, 0, -o_x], [0, 1, -o_y], [0, 0, 1]])
  transform_matrix = np.dot(np.dot(offset_matrix, matrix), reset_matrix)
  return transform_matrix


def rotate(x, rg, fill_mode='nearest'):
  """Performs a random rotation of a Numpy image tensor.

  # Arguments
      x: Input tensor. Must be 3D.
      rg: Rotation range, in degrees.
          can be negative or positive
      fill_mode: Points outside the boundaries of the input
          are filled according to the given mode
          (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).

  # Returns
      Rotated Numpy image tensor.
  """
  theta = np.pi / 180 * rg
  rotation_matrix = np.array([[np.cos(theta), -np.sin(theta), 0],
                              [np.sin(theta), np.cos(theta), 0],
                              [0, 0, 1]])

  h, w = x.shape[0], x.shape[1]
  transform_matrix = transform_matrix_offset_center(rotation_matrix, h, w)
  x = apply_transform(x, transform_matrix, fill_mode)
  return x


def shift(x, wrg, hrg, fill_mode='nearest'):
  """Performs a random spatial shift of a Numpy image tensor.

  # Arguments
      x: Input tensor. Must be 3D.
      wrg: Width shift range, as a float fraction of the width.
          can be negative or positive
      hrg: Height shift range, as a float fraction of the height.
          can be negative or positive
      fill_mode: Points outside the boundaries of the input
          are filled according to the given mode
          (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).

  # Returns
      Shifted Numpy image tensor.
  """
  h, w = x.shape[0], x.shape[1]
  tx = hrg * h
  ty = wrg * w
  translation_matrix = np.array([[1, 0, tx],
                                 [0, 1, ty],
                                 [0, 0, 1]])

  transform_matrix = translation_matrix  # no need to do offset
  x = apply_transform(x, transform_matrix, fill_mode)
  return x


def zoom(x, zoom_width, zoom_height, fill_mode='nearest'):
  """Performs a random spatial zoom of a Numpy image tensor.

  # Arguments
      x: Input tensor. Must be 3D.
      zoom_range: Tuple of floats; zoom range for width and height.
      fill_mode: Points outside the boundaries of the input
          are filled according to the given mode
          (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).

  # Returns
      Zoomed Numpy image tensor.

  # Raises
      ValueError: if `zoom_range` isn't a tuple.
  """
  if zoom_width == 1 and zoom_height == 1:
    zx, zy = 1, 1
  else:
    zx, zy = np.random.uniform(zoom_width, zoom_height, 2)
  zoom_matrix = np.array([[zx, 0, 0],
                          [0, zy, 0],
                          [0, 0, 1]])

  h, w = x.shape[0], x.shape[1]
  transform_matrix = transform_matrix_offset_center(zoom_matrix, h, w)
  x = apply_transform(x, transform_matrix, fill_mode)
  return x


def shear(x, intensity, fill_mode='nearest'):
  """Performs a random spatial shear of a Numpy image tensor.

  # Arguments
      x: Input tensor. Must be 3D.
      intensity: Transformation intensity.
          can be negative or positive
      fill_mode: Points outside the boundaries of the input
          are filled according to the given mode
          (one of `{'constant', 'nearest', 'reflect', 'wrap'}`).
      cval: Value used for points outside the boundaries
          of the input if `mode='constant'`.

  # Returns
      Sheared Numpy image tensor.
  """
  shear = intensity
  shear_matrix = np.array([[1, -np.sin(shear), 0],
                           [0, np.cos(shear), 0],
                           [0, 0, 1]])

  h, w = x.shape[0], x.shape[1]
  transform_matrix = transform_matrix_offset_center(shear_matrix, h, w)
  x = apply_transform(x, transform_matrix, fill_mode)
  return x


def one_hot(y, nb_classes):
  a = np.zeros((len(y), nb_classes), 'uint8')
  a[np.arange(len(y)), y] = 1
  return a


def read(path, grayscale=False, crop=None, scale=None, target_size=None,
         transpose=None, resample_mode=2):
  """
  Parameters
  ----------
  grayscale: bool
      force to convert Image to grayscale or not
  crop: 4-tuple of int
       (left, upper, right, lower)
  scale: 2-tuple of float
      (factor of width, factor of height) of output images
  target_size: 2-tuple of int
      desire size for image (image will be padded if the size
      mis-match)
  transpose: int, or list of int
      if a list of int is provided, will return a list of images
      <0: Do nothing
      0: PIL.Image.FLIP_LEFT_RIGHT
      1: PIL.Image.FLIP_TOP_BOTTOM
      2: PIL.Image.ROTATE_90
      3: PIL.Image.ROTATE_180
      4: PIL.Image.ROTATE_270
      5: PIL.Image.TRANSPOSE
  resample_mode: int
      0 = PIL.Image.NEAREST: use nearest neighbour
      1 = PIL.Image.LANCZOS: a high-quality downsampling filter
      2 = PIL.Image.BILINEAR: linear interpolation
      3 = PIL.Image.BICUBIC: cubic spline interpolation
  Return
  ------
  image: (height, width, channel) if RGB image
         (height, width) if Grayscale image

  Example
  -------
  >>> x = image.read('/Users/trungnt13/tmp/1090.jpg', grayscale=True,
  ...                scale=0.4, target_size=(800, 900),
  ...                transpose=(-1, 0, 1, 2, 3, 4))
  ... # return a list of 6 images with size=(800, 900)
  >>> print([i.shape for i in x])
  ... # [(900, 800), (900, 800), (900, 800), (900, 800), (900, 800), (900, 800)]
  >>> plt.subplot(1, 6, 1)
  >>> plt.imshow(x[0], cmap='gray')
  >>> plt.subplot(1, 6, 2)
  >>> plt.imshow(x[1], cmap='gray')
  >>> plt.subplot(1, 6, 3)
  >>> plt.imshow(x[2], cmap='gray')
  >>> plt.subplot(1, 6, 4)
  >>> plt.imshow(x[3], cmap='gray')
  >>> plt.subplot(1, 6, 5)
  >>> plt.imshow(x[4], cmap='gray')
  >>> plt.subplot(1, 6, 6)
  >>> plt.imshow(x[5], cmap='gray')
  >>> plt.show(block=True)

  Benchmarks
  ----------
  RESAMPLE_MODE = ANTIALIAS:
   * scale and target_size: 0.019858 (sec)
   * Only scale: 0.009354 (sec)
   * NO scale or target_size: 0.002620 (sec)
  RESAMPLE_MODE = BILINEAR:
   * scale and target_size: 0.012858 (sec)
   * Only scale: 0.006643 (sec)
   * NO scale or target_size: 0.002419 (sec)
  """
  img = Image.open(path, mode="r")
  if grayscale:
    img = img.convert('L')
  else:  # Ensure 3 channel even when loaded image is grayscale
    img = img.convert('RGB')
  # ====== crop ====== #
  if crop:
    img = img.crop(crop)
  # ====== scale ====== #
  if scale and scale != 1.:
    size = img.size
    scale = as_tuple(scale, len(size), t=float)
    scale = tuple([int(i * j) for i, j in zip(scale, size)])
    img = img.resize(scale, resample=resample_mode)
  # ====== return X ====== #
  X = []
  transpose = (-1,) if transpose is None else as_tuple(transpose, t=int)
  # ====== traverse each transposed images ====== #
  orig_img = img
  for t in transpose:
    img = orig_img.transpose(t) if t >= 0 else orig_img
    # ====== target_size ====== #
    if target_size:
      size = img.size
      target_size = as_tuple(target_size, 2, int)
      # image is bigger than the target size
      scale = [j / i for i, j in zip(size, target_size)]
      # print(size, target_size, scale)
      # resize smaller or bigger and preserve the ratio
      if any(i < 1 for i in scale) or all(i > 1 for i in scale):
        scale = min(scale)
        scale = [int(round(scale * i)) for i in size]
        img = img.resize(scale, resample=resample_mode)
      # centerlize the image
      size = img.size
      if size != target_size:
        new_img = Image.new('L' if grayscale else 'RGB',
                            target_size,   # A4 at 72dpi
                            0 if grayscale else (0, 0, 0))  # White
        x = 0 if size[0] == target_size[0] else (target_size[0] - size[0]) // 2
        y = 0 if size[1] == target_size[1] else (target_size[1] - size[1]) // 2
        new_img.paste(img, (x, y))
        img.close()
        img = new_img
    X.append(np.asarray(img))
    img.close()
  # ====== end ====== #
  orig_img.close()
  # X = [x.transpose(-1, 0, 1) if x.ndim == 3 else x for x in X]
  return X[0] if len(X) == 1 else X
